# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable
from functools import lru_cache

import numpy as np


def verify_quantum_operator_matrix_dimensions(matrix: np.ndarray) -> None:
    """Verifies matrix is square and matrix dimensions are positive powers of 2,
    raising `ValueError` otherwise.

    Args:
        matrix (ndarray): matrix to verify

    Raises:
        ValueError: If `matrix` is not a two-dimensional square matrix,
            or has a dimension length that is not a positive power of 2
    """
    if not is_square_matrix(matrix):
        raise ValueError(f"{matrix} is not a two-dimensional square matrix")

    matrix = np.array(matrix, dtype=complex)
    qubit_count = int(np.log2(matrix.shape[0]))

    if 2**qubit_count != matrix.shape[0] or qubit_count < 1:
        raise ValueError(f"`matrix` dimension {matrix.shape[0]} is not a positive power of 2")


def is_hermitian(matrix: np.ndarray) -> bool:
    r"""Whether matrix is Hermitian

    A square matrix :math:`U` is Hermitian if

    .. math:: U = U^\dagger

    where :math:`U^\dagger` is the conjugate transpose of :math:`U`.

    Args:
        matrix (ndarray): matrix to verify

    Returns:
        bool: If matrix is Hermitian
    """
    return np.allclose(matrix, matrix.conj().T)


def is_square_matrix(matrix: np.ndarray) -> bool:
    """Whether matrix is square, meaning it has exactly two dimensions and the dimensions are equal

    Args:
        matrix (np.ndarray): matrix to verify

    Returns:
        bool: If matrix is square
    """
    return len(matrix.shape) == 2 and matrix.shape[0] == matrix.shape[1]


def is_unitary(matrix: np.ndarray) -> bool:
    r"""Whether matrix is unitary

    A square matrix :math:`U` is unitary if

    .. math:: UU^\dagger = I

    where :math:`U^\dagger` is the conjugate transpose of :math:`U`
    and :math:`I` is the identity matrix.

    Args:
        matrix (np.ndarray): matrix to verify

    Returns:
        bool: If matrix is unitary
    """
    return np.allclose(np.eye(len(matrix)), matrix.dot(matrix.T.conj()))


def is_cptp(matrices: Iterable[np.ndarray]) -> bool:
    """Whether a transformation defined by these matrics as Kraus operators is a
    completely positive trace preserving (CPTP) map. This is the requirement for
    a transformation to be a quantum channel.
    Reference: Section 8.2.3 in Nielsen & Chuang (2010) 10th edition.

    Args:
        matrices (Iterable[ndarray]): List of matrices representing Kraus operators.

    Returns:
        bool: If the matrices define a CPTP map.
    """
    E = sum([np.dot(matrix.T.conjugate(), matrix) for matrix in matrices])
    return np.allclose(E, np.eye(*E.shape))


@lru_cache
def get_pauli_eigenvalues(num_qubits: int) -> np.ndarray:
    """Get the eigenvalues of Pauli operators and their tensor products as
    an immutable Numpy ndarray.

    Args:
        num_qubits (int): the number of qubits the operator acts on

    Returns:
        np.ndarray: the eigenvalues of a Pauli product operator of the given size
    """
    if num_qubits == 1:
        eigs = np.array([1, -1])
        eigs.setflags(write=False)
        return eigs
    eigs = np.concatenate(
        [get_pauli_eigenvalues(num_qubits - 1), -get_pauli_eigenvalues(num_qubits - 1)]
    )
    eigs.setflags(write=False)
    return eigs



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from braket.circuits.operator import Operator
from braket.circuits.serialization import IRType, SerializationProperties
from braket.registers.qubit_set import QubitSet


class CompilerDirective(Operator):
    """A directive specifying how the compiler should process a part of the circuit.

    For example, a directive may tell the compiler not to modify some gates in the circuit.
    """

    def __init__(self, ascii_symbols: Sequence[str]):
        """Inits a `CompilerDirective`.

        Args:
            ascii_symbols (Sequence[str]): ASCII string symbols for the compiler directiver.
                These are used when printing a diagram of circuits.
        """
        if ascii_symbols is None:
            raise ValueError("ascii_symbols must not be None")
        self._ascii_symbols = tuple(ascii_symbols)

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        """tuple[str, ...]: Returns the ascii symbols for the compiler directive."""
        return self._ascii_symbols

    def to_ir(
        self,
        target: QubitSet | None = None,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: SerializationProperties | None = None,
        **kwargs,
    ) -> Any:
        """Returns IR object of the compiler directive.

        Args:
            target (QubitSet | None): target qubit(s). Defaults to None
            ir_type(IRType) : The IRType to use for converting the compiler directive object to its
                IR representation. Defaults to IRType.JAQCD.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Any: IR object of the compiler directive.

        Raises:
            ValueError: If the supplied `ir_type` is not supported.
        """
        if ir_type == IRType.JAQCD:
            return self._to_jaqcd()
        elif ir_type == IRType.OPENQASM:
            return self._to_openqasm()
        else:
            raise ValueError(f"Supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self) -> Any:
        """Returns the JAQCD representation of the compiler directive."""
        raise NotImplementedError("to_jaqcd has not been implemented yet.")

    def _to_openqasm(self) -> str:
        """Returns the openqasm string representation of the compiler directive."""
        raise NotImplementedError("to_openqasm has not been implemented yet.")

    def counterpart(self) -> CompilerDirective:
        """Returns the "opposite" counterpart to this compiler directive.

        For example, the counterpart of a directive that starts a box
        is the directive that ends the box.

        Returns:
            CompilerDirective: The counterpart compiler directive
        """
        raise NotImplementedError(
            f"Compiler directive {self.name} does not have counterpart implemented"
        )

    def __eq__(self, other: CompilerDirective):
        return isinstance(other, CompilerDirective) and self.name == other.name

    def __repr__(self):
        return self.name



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import time
from typing import Optional, Union


def log_metric(
    metric_name: str,
    value: Union[float, int],
    timestamp: Optional[float] = None,
    iteration_number: Optional[int] = None,
) -> None:
    """Records Braket Hybrid Job metrics.

    Args:
        metric_name (str): The name of the metric.

        value (Union[float, int]): The value of the metric.

        timestamp (Optional[float]): The time the metric data was received, expressed
            as the number of seconds since the epoch. Default: Current system time.

        iteration_number (Optional[int]): The iteration number of the metric.
    """
    logged_timestamp = timestamp or time.time()
    metric_list = [f"Metrics - timestamp={logged_timestamp}; {metric_name}={value};"]
    if iteration_number is not None:
        metric_list.append(f" iteration_number={iteration_number};")
    metric_line = "".join(metric_list)
    print(metric_line)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import functools
import itertools
import math
import numbers
from copy import deepcopy
from typing import ClassVar, Union

import numpy as np

from braket.circuits.gate import Gate
from braket.circuits.observable import Observable, StandardObservable
from braket.circuits.quantum_operator_helpers import (
    get_pauli_eigenvalues,
    is_hermitian,
    verify_quantum_operator_matrix_dimensions,
)
from braket.circuits.serialization import IRType, OpenQASMSerializationProperties
from braket.registers.qubit_set import QubitSet


class H(StandardObservable):
    """Hadamard operation as an observable."""

    def __init__(self):
        """Examples:
        >>> Observable.H()
        """
        super().__init__(ascii_symbols=["H"])

    def _unscaled(self) -> StandardObservable:
        return H()

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return ["h"]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = serialization_properties.format_target(int(target[0]))
            return f"{coef_prefix}h({qubit_target})"
        else:
            return f"{coef_prefix}h all"

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * (
            1.0 / np.sqrt(2.0) * np.array([[1.0, 1.0], [1.0, -1.0]], dtype=complex)
        )

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return tuple([Gate.Ry(-math.pi / 4)])  # noqa: C409


Observable.register_observable(H)


class I(Observable):  # noqa: E742
    """Identity operation as an observable."""

    def __init__(self):
        """Examples:
        >>> Observable.I()
        """
        super().__init__(qubit_count=1, ascii_symbols=["I"])

    def _unscaled(self) -> Observable:
        return I()

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return ["i"]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = serialization_properties.format_target(int(target[0]))
            return f"{coef_prefix}i({qubit_target})"
        else:
            return f"{coef_prefix}i all"

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * np.eye(2, dtype=complex)

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return ()

    @property
    def eigenvalues(self) -> np.ndarray:
        """Returns the eigenvalues of this observable.

        Returns:
            np.ndarray: The eigenvalues of this observable.
        """
        return self.coefficient * np.ones(2)

    def eigenvalue(self, index: int) -> float:
        return self.coefficient * 1.0


Observable.register_observable(I)


class X(StandardObservable):
    """Pauli-X operation as an observable."""

    def __init__(self):
        """Examples:
        >>> Observable.X()
        """
        super().__init__(ascii_symbols=["X"])

    def _unscaled(self) -> StandardObservable:
        return X()

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return ["x"]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = serialization_properties.format_target(int(target[0]))
            return f"{coef_prefix}x({qubit_target})"
        else:
            return f"{coef_prefix}x all"

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * np.array([[0.0, 1.0], [1.0, 0.0]], dtype=complex)

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return tuple([Gate.H()])  # noqa: C409


Observable.register_observable(X)


class Y(StandardObservable):
    """Pauli-Y operation as an observable."""

    def __init__(self):
        """Examples:
        >>> Observable.Y()
        """
        super().__init__(ascii_symbols=["Y"])

    def _unscaled(self) -> StandardObservable:
        return Y()

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return ["y"]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = serialization_properties.format_target(int(target[0]))
            return f"{coef_prefix}y({qubit_target})"
        else:
            return f"{coef_prefix}y all"

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * np.array([[0.0, -1.0j], [1.0j, 0.0]], dtype=complex)

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return tuple([Gate.Z(), Gate.S(), Gate.H()])  # noqa: C409


Observable.register_observable(Y)


class Z(StandardObservable):
    """Pauli-Z operation as an observable."""

    def __init__(self):
        """Examples:
        >>> Observable.Z()
        """
        super().__init__(ascii_symbols=["Z"])

    def _unscaled(self) -> StandardObservable:
        return Z()

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return ["z"]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = serialization_properties.format_target(int(target[0]))
            return f"{coef_prefix}z({qubit_target})"
        else:
            return f"{coef_prefix}z all"

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * np.array([[1.0, 0.0], [0.0, -1.0]], dtype=complex)

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return ()


Observable.register_observable(Z)


class TensorProduct(Observable):
    """Tensor product of observables"""

    def __init__(self, observables: list[Observable]):
        """Initializes a `TensorProduct`.

        Args:
            observables (list[Observable]): List of observables for tensor product

        Examples:
            >>> t1 = Observable.Y() @ Observable.X()
            >>> t1.to_matrix()
            array([[0.+0.j, 0.+0.j, 0.-0.j, 0.-1.j],
            [0.+0.j, 0.+0.j, 0.-1.j, 0.-0.j],
            [0.+0.j, 0.+1.j, 0.+0.j, 0.+0.j],
            [0.+1.j, 0.+0.j, 0.+0.j, 0.+0.j]])
            >>> t2 = Observable.Z() @ t1
            >>> t2.factors
            (Z('qubit_count': 1), Y('qubit_count': 1), X('qubit_count': 1))

        Note: You must provide the list of observables for the tensor product to be evaluated
        in the order that you want the tensor product to be calculated.
        For `TensorProduct(observables=[ob1, ob2, ob3])`, the tensor product's matrix is the
        result of the tensor product of `ob1`, `ob2`, `ob3`, or `np.kron(np.kron(ob1.to_matrix(),
        ob2.to_matrix()), ob3.to_matrix())`.
        """
        flattened_observables = []
        for obs in observables:
            if isinstance(obs, TensorProduct):
                for nested_obs in obs.factors:
                    flattened_observables.append(nested_obs)
                # make sure you don't lose coefficient of tensor product
                flattened_observables[-1] *= obs.coefficient
            elif isinstance(obs, Sum):
                raise TypeError("Sum observables not allowed in TensorProduct")
            else:
                flattened_observables.append(obs)
        qubit_count = sum([obs.qubit_count for obs in flattened_observables])
        # aggregate all coefficients for the product, since aX @ bY == ab * X @ Y
        coefficient = np.prod([obs.coefficient for obs in flattened_observables])
        unscaled_factors = tuple(obs._unscaled() for obs in flattened_observables)
        display_name = (
            f"{coefficient if coefficient != 1 else ''}"
            f"{'@'.join([obs.ascii_symbols[0] for obs in unscaled_factors])}"
        )
        super().__init__(qubit_count=qubit_count, ascii_symbols=[display_name] * qubit_count)
        self._coef = coefficient
        self._factors = unscaled_factors
        self._factor_dimensions = tuple(
            len(factor.to_matrix()) for factor in reversed(self._factors)
        )
        self._eigenvalue_indices = {}
        self._all_eigenvalues = None

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        return tuple(
            f"{self.coefficient if self.coefficient != 1 else ''}"
            f"{'@'.join([obs.ascii_symbols[0] for obs in self.factors])}"
            for _ in range(self.qubit_count)
        )

    def _unscaled(self) -> Observable:
        copied = TensorProduct(observables=self.factors)
        copied._coef = 1
        return copied

    def _to_jaqcd(self) -> list[str]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        ir = []
        for obs in self.factors:
            ir.extend(obs.to_ir())
        return ir

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        factors = []
        use_qubits = iter(target)
        for obs in self._factors:
            obs_target = QubitSet()
            num_qubits = int(np.log2(obs.to_matrix().shape[0]))
            for _ in range(num_qubits):
                obs_target.add(next(use_qubits))
            factors.append(
                obs.to_ir(
                    target=obs_target,
                    ir_type=IRType.OPENQASM,
                    serialization_properties=serialization_properties,
                )
            )
        return f"{coef_prefix}{' @ '.join(factors)}"

    @property
    def factors(self) -> tuple[Observable, ...]:
        """tuple[Observable]: The observables that comprise this tensor product."""
        return self._factors

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * functools.reduce(
            np.kron, [obs.to_matrix() for obs in self.factors]
        )

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        """Returns the basis rotation gates for this observable.

        Returns:
            tuple[Gate, ...]: The basis rotation gates for this observable.
        """
        gates = []
        for obs in self.factors:
            gates.extend(obs.basis_rotation_gates)
        return tuple(gates)

    @property
    def eigenvalues(self) -> np.ndarray:
        """Returns the eigenvalues of this observable.

        Returns:
            np.ndarray: The eigenvalues of this observable.
        """
        if self._all_eigenvalues is None:
            self._all_eigenvalues = TensorProduct._compute_eigenvalues(
                self._factors, self.qubit_count
            )
        return self.coefficient * self._all_eigenvalues

    def eigenvalue(self, index: int) -> float:
        """Returns the eigenvalue of this observable at the given index.

        The eigenvalues are ordered by their corresponding computational basis state
        after diagonalization.

        Args:
            index (int): The index of the desired eigenvalue

        Returns:
            float: The `index` th eigenvalue of the observable.
        """
        if index in self._eigenvalue_indices:
            return self._eigenvalue_indices[index]
        dimension = 2**self.qubit_count
        if index >= dimension:
            raise ValueError(
                f"Index {index} requested but observable has at most {dimension} eigenvalues"
            )
        # Calculating the eigenvalue amounts to converting the index to a new heterogeneous base
        # and multiplying the eigenvalues of each factor at the corresponding digit
        product = 1
        quotient = index
        for i in range(len(self._factors)):
            quotient, remainder = divmod(quotient, self._factor_dimensions[i])
            product *= self._factors[-i - 1].eigenvalue(remainder)
        self._eigenvalue_indices[index] = product
        return self.coefficient * self._eigenvalue_indices[index]

    def __repr__(self):
        return "TensorProduct(" + ", ".join([repr(o) for o in self.factors]) + ")"

    def __eq__(self, other: TensorProduct):
        return self.matrix_equivalence(other)

    @staticmethod
    def _compute_eigenvalues(observables: tuple[Observable], num_qubits: int) -> np.ndarray:
        if False in [isinstance(observable, StandardObservable) for observable in observables]:
            # Tensor product of observables contains a mixture
            # of standard and non-standard observables
            eigenvalues = np.array([1])
            for k, g in itertools.groupby(observables, lambda x: isinstance(x, StandardObservable)):
                if k:
                    # Subgroup g contains only standard observables.
                    eigenvalues = np.kron(eigenvalues, get_pauli_eigenvalues(len(list(g))))
                else:
                    # Subgroup g contains only non-standard observables.
                    for nonstandard in g:
                        # loop through all non-standard observables
                        eigenvalues = np.kron(eigenvalues, nonstandard.eigenvalues)
        else:
            eigenvalues = get_pauli_eigenvalues(num_qubits=num_qubits)

        eigenvalues.setflags(write=False)
        return eigenvalues


Observable.register_observable(TensorProduct)


class Sum(Observable):
    """Sum of observables"""

    def __init__(self, observables: list[Observable], display_name: str = "Hamiltonian"):
        """Inits a `Sum`.

        Args:
            observables (list[Observable]): List of observables for Sum
            display_name (str): Name to use for an instance of this Sum
                observable for circuit diagrams. Defaults to `Hamiltonian`.

        Examples:
            >>> t1 = -3 * Observable.Y() + 2 * Observable.X()
            Sum(X('qubit_count': 1), Y('qubit_count': 1))
            >>> t1.summands
            (X('qubit_count': 1), Y('qubit_count': 1))
        """
        flattened_observables = []
        for obs in observables:
            if isinstance(obs, Sum):
                for nested_obs in obs.summands:
                    flattened_observables.append(nested_obs)
            else:
                flattened_observables.append(obs)

        self._summands = tuple(flattened_observables)
        qubit_count = max(flattened_observables, key=lambda obs: obs.qubit_count).qubit_count
        super().__init__(qubit_count=qubit_count, ascii_symbols=[display_name] * qubit_count)

    def __mul__(self, other: numbers.Number) -> Observable:
        """Scalar multiplication"""
        if isinstance(other, numbers.Number):
            sum_copy = deepcopy(self)
            for i, _obs in enumerate(sum_copy.summands):
                sum_copy._summands[i]._coef *= other
            return sum_copy
        raise TypeError("Observable coefficients must be numbers.")

    def _to_jaqcd(self) -> list[str]:
        raise NotImplementedError("Sum Observable is not supported in Jaqcd")

    def _to_openqasm(
        self,
        serialization_properties: OpenQASMSerializationProperties,
        target: list[QubitSet] = None,
    ) -> str:
        if len(self.summands) != len(target):
            raise ValueError(
                f"Invalid target of length {len(target)} for Sum with {len(self.summands)} terms"
            )
        for i, (term, term_target) in enumerate(zip(self.summands, target)):
            if term.qubit_count != len(term_target):
                raise ValueError(
                    f"Invalid target for term {i} of Sum. "
                    f"Expected {term.qubit_count} targets, got {len(term_target)}"
                )
        return " + ".join(
            obs.to_ir(
                target=term_target,
                ir_type=IRType.OPENQASM,
                serialization_properties=serialization_properties,
            )
            for obs, term_target in zip(self.summands, target)
        ).replace("+ -", "- ")

    @property
    def summands(self) -> tuple[Observable, ...]:
        """tuple[Observable]: The observables that comprise this sum."""
        return self._summands

    def to_matrix(self) -> np.ndarray:
        raise NotImplementedError("Matrix operation is not supported for Sum")

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        raise NotImplementedError("Basis rotation calculation not supported for Sum")

    @property
    def eigenvalues(self) -> np.ndarray:
        raise NotImplementedError("Eigenvalue calculation not supported for Sum")

    def eigenvalue(self, index: int) -> float:
        raise NotImplementedError("Eigenvalue calculation not supported for Sum")

    def __repr__(self):
        return "Sum(" + ", ".join([repr(o) for o in self.summands]) + ")"

    def __eq__(self, other: Sum):
        return repr(self) == repr(other)

    @staticmethod
    def _compute_eigenvalues(observables: tuple[Observable], num_qubits: int) -> np.ndarray:
        raise NotImplementedError("Eigenvalue calculation not supported for Sum")


Observable.register_observable(Sum)


class Hermitian(Observable):
    """Hermitian matrix as an observable."""

    # Cache of eigenpairs
    _eigenpairs: ClassVar = {}

    def __init__(self, matrix: np.ndarray, display_name: str = "Hermitian"):
        """Inits a `Hermitian`.

        Args:
            matrix (np.ndarray): Hermitian matrix that defines the observable.
            display_name (str): Name to use for an instance of this Hermitian matrix
                observable for circuit diagrams. Defaults to `Hermitian`.

        Raises:
            ValueError: If `matrix` is not a two-dimensional square matrix,
                or has a dimension length that is not a positive power of 2,
                or is not Hermitian.

        Examples:
            >>> Observable.Hermitian(matrix=np.array([[0, 1],[1, 0]]))
        """
        verify_quantum_operator_matrix_dimensions(matrix)
        self._matrix = np.array(matrix, dtype=complex)
        if not is_hermitian(self._matrix):
            raise ValueError(f"{self._matrix} is not hermitian")

        qubit_count = int(np.log2(self._matrix.shape[0]))
        eigendecomposition = Hermitian._get_eigendecomposition(self._matrix)
        self._eigenvalues = eigendecomposition["eigenvalues"]
        self._diagonalizing_gates = (
            Gate.Unitary(matrix=eigendecomposition["eigenvectors"].conj().T),
        )

        super().__init__(qubit_count=qubit_count, ascii_symbols=[display_name] * qubit_count)

    def _unscaled(self) -> Observable:
        return Hermitian(matrix=self._matrix, display_name=self.ascii_symbols[0])

    def _to_jaqcd(self) -> list[list[list[list[float]]]]:
        if self.coefficient != 1:
            raise ValueError("Observable coefficients not supported with Jaqcd")
        return [
            [[[element.real, element.imag] for element in row] for row in self._matrix.tolist()]
        ]

    def _to_openqasm(
        self, serialization_properties: OpenQASMSerializationProperties, target: QubitSet = None
    ) -> str:
        coef_prefix = f"{self.coefficient} * " if self.coefficient != 1 else ""
        if target:
            qubit_target = ", ".join(
                [serialization_properties.format_target(int(t)) for t in target]
            )
            return (
                f"{coef_prefix}"
                f"hermitian({self._serialized_matrix_openqasm_matrix()}) {qubit_target}"
            )
        else:
            return f"{coef_prefix}hermitian({self._serialized_matrix_openqasm_matrix()}) all"

    def _serialized_matrix_openqasm_matrix(self) -> str:
        serialized = str([[f"{complex(elem)}" for elem in row] for row in self._matrix.tolist()])
        for replacements in [("(", ""), (")", ""), ("'", ""), ("j", "im")]:
            serialized = serialized.replace(replacements[0], replacements[1])
        return serialized

    def to_matrix(self) -> np.ndarray:
        return self.coefficient * self._matrix

    def __eq__(self, other: Hermitian) -> bool:
        return self.matrix_equivalence(other)

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        return self._diagonalizing_gates

    @property
    def eigenvalues(self) -> np.ndarray:
        """Returns the eigenvalues of this observable.

        Returns:
            np.ndarray: The eigenvalues of this observable.
        """
        return self._eigenvalues

    def eigenvalue(self, index: int) -> float:
        return self._eigenvalues[index]

    @staticmethod
    def _get_eigendecomposition(matrix: np.ndarray) -> dict[str, np.ndarray]:
        """Decomposes the Hermitian matrix into its eigenvectors and associated eigenvalues.
        The eigendecomposition is cached so that if another Hermitian observable
        is created with the same matrix, the eigendecomposition doesn't have to
        be recalculated.

        Args:
            matrix (ndarray): The Hermitian matrix.

        Returns:
            dict[str, ndarray]: The keys are "eigenvectors_conj_t", mapping to the
            conjugate transpose of a matrix whose columns are the eigenvectors of the matrix,
            and "eigenvalues", a list of associated eigenvalues in the order of their
            corresponding eigenvectors in the "eigenvectors" matrix. These cached values
            are immutable.
        """
        mat_key = tuple(matrix.flatten().tolist())
        if mat_key not in Hermitian._eigenpairs:
            eigenvalues, eigenvectors = np.linalg.eigh(matrix)
            eigenvalues.setflags(write=False)
            Hermitian._eigenpairs[mat_key] = {
                "eigenvectors": eigenvectors,
                "eigenvalues": eigenvalues,
            }
        return Hermitian._eigenpairs[mat_key]

    def __repr__(self):
        matrix_str = np.array2string(self.to_matrix()).replace("\n", ",")
        return f"{self.name}('qubit_count': {self.qubit_count}, 'matrix': {matrix_str})"


Observable.register_observable(Hermitian)


def observable_from_ir(ir_observable: list[Union[str, list[list[list[float]]]]]) -> Observable:
    """Create an observable from the IR observable list. This can be a tensor product of
    observables or a single observable.

    Args:
        ir_observable (list[Union[str, list[list[list[float]]]]]): observable as defined in IR

    Returns:
        Observable: observable object
    """
    if len(ir_observable) == 1:
        return _observable_from_ir_list_item(ir_observable[0])
    else:
        observable = TensorProduct([_observable_from_ir_list_item(obs) for obs in ir_observable])
        return observable


def _observable_from_ir_list_item(observable: Union[str, list[list[list[float]]]]) -> Observable:
    if observable == "i":
        return I()
    elif observable == "h":
        return H()
    elif observable == "x":
        return X()
    elif observable == "y":
        return Y()
    elif observable == "z":
        return Z()
    else:
        try:
            matrix = np.array(
                [[complex(element[0], element[1]) for element in row] for row in observable]
            )
            return Hermitian(matrix)
        except Exception as e:
            raise ValueError(f"Invalid observable specified: {observable} error: {e}")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable
from typing import Any, Optional, Union

from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.circuits.noise_model.circuit_instruction_criteria import CircuitInstructionCriteria
from braket.circuits.noise_model.criteria import Criteria, CriteriaKey, CriteriaKeyResult
from braket.circuits.noise_model.criteria_input_parsing import (
    parse_operator_input,
    parse_qubit_input,
)
from braket.registers.qubit_set import QubitSetInput


class GateCriteria(CircuitInstructionCriteria):
    """This class models noise Criteria based on named Braket SDK Gates."""

    def __init__(
        self,
        gates: Optional[Union[Gate, Iterable[Gate]]] = None,
        qubits: Optional[QubitSetInput] = None,
    ):
        """Creates Gate-based Criteria. See instruction_matches() for more details.

        Args:
            gates (Optional[Union[Gate, Iterable[Gate]]]): A set of relevant Gates. All the Gates
                must have the same fixed_qubit_count(). Optional. If gates are not provided
                this matcher will match on all gates.
            qubits (Optional[QubitSetInput]): A set of relevant qubits. If no qubits
                are provided, all (possible) qubits are considered to be relevant.

        Raises:
            ValueError: If the gates don't all operate on the same number of qubits, or if
            qubits are not valid targets for the provided gates.
        """
        self._gates = parse_operator_input(gates)
        expected_qubit_count = next(iter(self._gates)).fixed_qubit_count() if self._gates else 0
        self._qubits = parse_qubit_input(qubits, expected_qubit_count)

    def __str__(self):
        gate_names = {gate.__name__ for gate in self._gates} if self._gates is not None else None
        return f"{self.__class__.__name__}({gate_names}, {self._qubits})"

    def __repr__(self):
        gate_names = {gate.__name__ for gate in self._gates} if self._gates is not None else None
        return f"{self.__class__.__name__}(gates={gate_names}, qubits={self._qubits})"

    def applicable_key_types(self) -> Iterable[CriteriaKey]:
        """Returns an Iterable of criteria keys.

        Returns:
            Iterable[CriteriaKey]: This Criteria operates on Gates and Qubits.
        """
        return [CriteriaKey.QUBIT, CriteriaKey.GATE]

    def get_keys(self, key_type: CriteriaKey) -> Union[CriteriaKeyResult, set[Any]]:
        """Gets the keys for a given CriteriaKey.

        Args:
            key_type (CriteriaKey): The relevant Criteria Key.

        Returns:
            Union[CriteriaKeyResult, set[Any]]: The return value is based on the key type:
            GATE will return a set of Gate classes that are relevant to this Criteria.
            QUBIT will return a set of qubit targets that are relevant to this Criteria, or
            CriteriaKeyResult.ALL if the Criteria is relevant for all (possible) qubits.
            All other keys will return an empty list.
        """
        if key_type == CriteriaKey.GATE:
            return CriteriaKeyResult.ALL if self._gates is None else self._gates
        if key_type == CriteriaKey.QUBIT:
            return CriteriaKeyResult.ALL if self._qubits is None else set(self._qubits)
        return set()

    def to_dict(self) -> dict:
        """Converts a dictionary representing an object of this class into an instance of this
        class.

        Returns:
            dict: A dictionary representing the serialized version of this Criteria.
        """
        qubits = list(self._qubits) if self._qubits is not None else None
        gates = [gate.__name__ for gate in self._gates] if self._gates is not None else None
        return {
            "__class__": self.__class__.__name__,
            "gates": gates,
            "qubits": qubits,
        }

    def instruction_matches(self, instruction: Instruction) -> bool:
        """Returns true if an Instruction matches the criteria.

        Args:
            instruction (Instruction): An Instruction to match.

        Returns:
            bool: Returns true if the operator is one of the Gates provided in the constructor and
            the target is a qubit (or set of qubits) provided in the constructor.
            If gates were not provided in the constructor, then this method will accept any Gate.
            If qubits were not provided in the constructor, then this method will accept any
            Instruction target.
        """
        if isinstance(instruction, Iterable):
            return False
        if not isinstance(instruction.operator, Gate):
            return False
        if self._gates is not None and type(instruction.operator) not in self._gates:
            return False
        return CircuitInstructionCriteria._check_target_in_qubits(self._qubits, instruction.target)

    @classmethod
    def from_dict(cls, criteria: dict) -> Criteria:
        """Deserializes a dictionary into a Criteria object.

        Args:
            criteria (dict): A dictionary representation of a GateCriteria.

        Returns:
            Criteria: A deserialized GateCriteria represented by the passed in
            serialized data.
        """
        gates = (
            [getattr(Gate, name) for name in criteria["gates"]]
            if criteria["gates"] is not None
            else None
        )
        return GateCriteria(gates, criteria["qubits"])


Criteria.register_criteria(GateCriteria)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock, call, patch

import pytest

from braket.jobs.metrics_data.cwl_metrics_fetcher import CwlMetricsFetcher


@pytest.fixture
def aws_session():
    _aws_session = Mock()
    return _aws_session


EXAMPLE_METRICS_LOG_LINES = [
    {
        "timestamp": "Test timestamp 0",
        "message": "Metrics - Test value 0",
    },
    {
        "timestamp": "Test timestamp 1",
        "message": "Metrics - Test value 1",
    },
    {
        "timestamp": "Test timestamp 2",
    },
    {
        "message": "Metrics - Test value 3",
    },
    {
        # This metrics fetcher will filter out log line that don't have a "Metrics -" tag.
        "message": "No prefix, Test value 4",
    },
]

EXPECTED_CALL_LIST = [
    call("Test timestamp 0", "Metrics - Test value 0"),
    call("Test timestamp 1", "Metrics - Test value 1"),
    call(None, "Metrics - Test value 3"),
]


@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.get_parsed_metrics")
@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_all_metrics_complete_results(mock_add_metrics, mock_get_metrics, aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.describe_log_streams.return_value = {
        "logStreams": [{"logStreamName": "stream name"}, {}]
    }
    logs_client_mock.get_log_events.return_value = {
        "events": EXAMPLE_METRICS_LOG_LINES,
        "nextForwardToken": None,
    }
    expected_result = {"Test": [0]}
    mock_get_metrics.return_value = expected_result

    fetcher = CwlMetricsFetcher(aws_session)
    result = fetcher.get_metrics_for_job("test_job")
    assert mock_add_metrics.call_args_list == EXPECTED_CALL_LIST
    assert result == expected_result


@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_log_streams_timeout(mock_add_metrics, aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.describe_log_streams.return_value = {
        "logStreams": [{"logStreamName": "stream name"}],
        "nextToken": "forever",
    }
    logs_client_mock.get_log_events.return_value = {
        "events": EXAMPLE_METRICS_LOG_LINES,
    }

    fetcher = CwlMetricsFetcher(aws_session, 0.1)
    result = fetcher.get_metrics_for_job("test_job")
    mock_add_metrics.assert_not_called()
    assert result == {}


@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_no_streams_returned(mock_add_metrics, aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.describe_log_streams.return_value = {}

    fetcher = CwlMetricsFetcher(aws_session)
    result = fetcher.get_metrics_for_job("test_job")
    logs_client_mock.describe_log_streams.assert_called()
    mock_add_metrics.assert_not_called()
    assert result == {}


@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.get_parsed_metrics")
@patch("braket.jobs.metrics_data.cwl_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_metrics_timeout(mock_add_metrics, mock_get_metrics, aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.describe_log_streams.return_value = {
        "logStreams": [{"logStreamName": "stream name"}]
    }
    logs_client_mock.get_log_events.side_effect = get_log_events_forever
    expected_result = {"Test": [0]}
    mock_get_metrics.return_value = expected_result

    fetcher = CwlMetricsFetcher(aws_session, 0.1)
    result = fetcher.get_metrics_for_job("test_job")
    logs_client_mock.get_log_events.assert_called()
    mock_add_metrics.assert_called()
    assert result == expected_result


def get_log_events_forever(*args, **kwargs):
    next_token = "1"
    token = kwargs.get("nextToken")
    if token and token == "1":
        next_token = "2"
    return {"events": EXAMPLE_METRICS_LOG_LINES, "nextForwardToken": next_token}



import os

"""
Run from the root of the git repository, will apply contents of header to the beginning of
all *.py files in the chosen directories. Script is idempotent, meaning it won't apply the
header to a file that already contains it.

Usage: python bin/apply-header.py
"""

HEADER = """# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""

ROOT_DIRS = ["src", "test", "."]


def main():
    for root_dir in ROOT_DIRS:
        for root, dirs, files in os.walk(root_dir):
            for py_file in python_files(files):
                idempotent_prepend(os.path.join(root, py_file), HEADER)

            # don't recurse "." directory, just look at local files
            if root_dir == ".":
                break


def python_files(files):
    return [file for file in files if file.endswith("py")]


def idempotent_prepend(filename: str, new_content: str) -> None:
    with open(filename, "r+") as file:
        existing_content = file.read()

        if existing_content.startswith(new_content):
            print(f"Skipping {filename}, already contains the content.")
        else:
            print(f"Applying content to {filename}...")
            file.seek(0, 0)
            file.write(new_content + existing_content)


if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import operator
from typing import Union

from openpulse import ast
from openqasm3.visitor import QASMTransformer
from oqpy.program import Program
from oqpy.timing import OQDurationLiteral


class _FreeParameterTransformer(QASMTransformer):
    """Walk the AST and evaluate FreeParameterExpressions."""

    def __init__(self, param_values: dict[str, float], program: Program):
        self.param_values = param_values
        self.program = program
        super().__init__()

    def visit_Identifier(
        self, identifier: ast.Identifier
    ) -> Union[ast.Identifier, ast.FloatLiteral]:
        """Visit an Identifier.

        If the Identifier is used to hold a `FreeParameterExpression`, it will be simplified
        using the given parameter values.

        Args:
            identifier (Identifier): The identifier.

        Returns:
            Union[Identifier, FloatLiteral]: The transformed identifier.
        """
        if identifier.name in self.param_values:
            return ast.FloatLiteral(float(self.param_values[identifier.name]))
        return identifier

    def visit_BinaryExpression(
        self, node: ast.BinaryExpression
    ) -> Union[ast.BinaryExpression, ast.FloatLiteral]:
        """Visit a BinaryExpression.

        Visit the operands and simplify if they are literals.

        Args:
            node (BinaryExpression): The node.

        Returns:
            Union[BinaryExpression, FloatLiteral]: The transformed identifier.
        """
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)
        ops = {
            ast.BinaryOperator["+"]: operator.add,
            ast.BinaryOperator["*"]: operator.mul,
            ast.BinaryOperator["**"]: operator.pow,
        }
        if isinstance(lhs, ast.FloatLiteral):
            if isinstance(rhs, ast.FloatLiteral):
                return ast.FloatLiteral(ops[node.op](lhs.value, rhs.value))
            elif isinstance(rhs, ast.DurationLiteral) and node.op == ast.BinaryOperator["*"]:
                return OQDurationLiteral(lhs.value * rhs.value).to_ast(self.program)
        return ast.BinaryExpression(op=node.op, lhs=lhs, rhs=rhs)

    def visit_UnaryExpression(
        self, node: ast.UnaryExpression
    ) -> Union[ast.UnaryExpression, ast.FloatLiteral]:
        """Visit an UnaryExpression.

        Visit the operand and simplify if it is a literal.

        Args:
            node (UnaryExpression): The node.

        Returns:
            Union[UnaryExpression, FloatLiteral]: The transformed identifier.
        """
        expression = self.visit(node.expression)
        if (
            isinstance(expression, (ast.FloatLiteral, ast.DurationLiteral))
            and node.op == ast.UnaryOperator["-"]
        ):
            return type(expression)(-expression.value)
        return ast.UnaryExpression(op=node.op, expression=node.expression)  # pragma: no cover



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any, Optional

from oqpy import PortVar
from oqpy.base import OQPyExpression


class Port:
    """Ports represent any input or output component meant to manipulate and observe qubits on
    a device. See https://openqasm.com/language/openpulse.html#ports for more details.
    """

    def __init__(self, port_id: str, dt: float, properties: Optional[dict[str, Any]] = None):
        """Initializes a Port.

        Args:
            port_id (str): str identifying a unique port on the device.
            dt (float): The smallest time step that may be used on the control hardware.
            properties (Optional[dict[str, Any]]): Dict containing properties of
                this port. Defaults to None.
        """
        self._port_id = port_id
        self._dt = dt
        self.properties = properties

    @property
    def id(self) -> str:
        """Returns a str indicating the port id."""
        return self._port_id

    @property
    def dt(self) -> float:
        """Returns the smallest time step that may be used on the control hardware."""
        return self._dt

    def __eq__(self, other: Port) -> bool:
        return self.id == other.id if isinstance(other, Port) else False

    def _to_oqpy_expression(self) -> OQPyExpression:
        return PortVar(name=self.id)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import base64
import subprocess
from pathlib import Path, PurePosixPath
from unittest.mock import Mock, patch

import pytest

from braket.jobs.local.local_job_container import _LocalJobContainer


@pytest.fixture
def repo_uri():
    return "012345678901.dkr.ecr.us-west-2.amazonaws.com"


@pytest.fixture
def image_uri(repo_uri):
    return f"{repo_uri}/my-repo:my-tag"


@pytest.fixture
def aws_session():
    _aws_session = Mock()
    return _aws_session


@pytest.fixture
def popen_result():
    mock = Mock()
    mock.stdout.readline.side_effect = [
        str.encode("this\n"),
        str.encode("is a\n"),
        str.encode("test\n"),
    ]
    mock.poll.side_effect = [None, None, 0]
    return mock


@patch("subprocess.check_output")
@patch("subprocess.run")
def test_start_and_stop(mock_run, mock_check_output, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
    ]
    with _LocalJobContainer(image_uri, aws_session):
        pass
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    assert mock_check_output.call_count == 2
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 1


@pytest.mark.parametrize(
    "forced_update, check_output, local_image_name",
    [
        (
            False,
            [
                str.encode(""),  # This means that the container image does not exist locally.
                str.encode("LocalImageName"),
                str.encode("RunningContainer"),
            ],
            "LocalImageName",
        ),
        (
            True,
            [
                str.encode(""),  # This means that the container image does not exist locally.
                str.encode("LocalImageName"),
                str.encode("RunningContainer"),
            ],
            "LocalImageName",
        ),
        (
            True,  # When force update is true, we'll pull containers, even if they exist locally.
            [
                str.encode("PreUpdateName"),  # This means that the container image exists locally.
                str.encode("PostUpdateName"),  # This means that the container image exists locally.
                str.encode("RunningContainer"),
            ],
            "PostUpdateName",
        ),
    ],
)
@patch("subprocess.check_output")
@patch("subprocess.run")
def test_pull_container(
    mock_run,
    mock_check_output,
    repo_uri,
    image_uri,
    aws_session,
    forced_update,
    check_output,
    local_image_name,
):
    running_container_name = "RunningContainer"
    test_token = "Test Token"
    mock_check_output.side_effect = check_output
    aws_session.ecr_client.get_authorization_token.return_value = {
        "authorizationData": [{"authorizationToken": base64.b64encode(str.encode(test_token))}]
    }
    with _LocalJobContainer(
        image_uri=image_uri, aws_session=aws_session, force_update=forced_update
    ):
        pass
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    assert mock_check_output.call_count == len(check_output)
    mock_run.assert_any_call(["docker", "login", "-u", "AWS", "-p", test_token, repo_uri])
    mock_run.assert_any_call(["docker", "pull", image_uri])
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 3


@patch("subprocess.check_output")
@patch("subprocess.run")
def test_pull_container_forced_update_invalid_name(
    mock_run, mock_check_output, repo_uri, aws_session
):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    mock_logger = Mock()
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
    ]
    with _LocalJobContainer(
        image_uri=local_image_name, aws_session=aws_session, force_update=True, logger=mock_logger
    ):
        pass
    mock_logger.warning.assert_called_with(f"Unable to update {local_image_name}.")
    mock_check_output.assert_any_call(["docker", "images", "-q", local_image_name])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    assert mock_check_output.call_count == 2
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 1


@patch("subprocess.check_output")
@patch("subprocess.run")
@patch("subprocess.Popen")
def test_run_job_success(
    mock_popen, mock_run, mock_check_output, repo_uri, image_uri, aws_session, popen_result
):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    env_variables = {
        "ENV0": "VALUE0",
        "ENV1": "VALUE1",
    }
    run_program_name = "Run Program Name"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(run_program_name),
    ]
    mock_popen.return_value = popen_result
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.run_local_job(env_variables)
        assert container.run_log == "this\nis a\ntest\n"
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    mock_check_output.assert_any_call(
        ["docker", "exec", running_container_name, "printenv", "SAGEMAKER_PROGRAM"]
    )
    assert mock_check_output.call_count == 3
    mock_popen.assert_called_with(
        [
            "docker",
            "exec",
            "-w",
            "/opt/ml/code/",
            "-e",
            "ENV0=VALUE0",
            "-e",
            "ENV1=VALUE1",
            running_container_name,
            "python",
            run_program_name,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    mock_run.assert_called_with(["docker", "stop", running_container_name])


@patch("subprocess.check_output")
@patch("subprocess.run")
@patch("subprocess.Popen")
def test_run_customer_script_fails(
    mock_popen, mock_run, mock_check_output, repo_uri, image_uri, aws_session, popen_result
):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    env_variables = {
        "ENV0": "VALUE0",
        "ENV1": "VALUE1",
    }
    run_program_name = "Run Program Name"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(run_program_name),
    ]
    popen_result.poll.side_effect = [None, None, 400]
    mock_popen.return_value = popen_result
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.run_local_job(env_variables)
        assert container.run_log == "this\nis a\ntest\nProcess exited with code: 400"
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    mock_check_output.assert_any_call(
        ["docker", "exec", running_container_name, "printenv", "SAGEMAKER_PROGRAM"]
    )
    assert mock_check_output.call_count == 3
    mock_popen.assert_called_with(
        [
            "docker",
            "exec",
            "-w",
            "/opt/ml/code/",
            "-e",
            "ENV0=VALUE0",
            "-e",
            "ENV1=VALUE1",
            running_container_name,
            "python",
            run_program_name,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    mock_run.assert_called_with(["docker", "stop", running_container_name])


@patch("subprocess.check_output")
@patch("subprocess.run")
@patch("subprocess.Popen")
def test_running_throws_exception(
    mock_popen, mock_run, mock_check_output, repo_uri, image_uri, aws_session, popen_result
):
    mock_logger = Mock()
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    env_variables = {
        "ENV0": "VALUE0",
        "ENV1": "VALUE1",
    }
    run_program_name = "Run Program Name"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(run_program_name),
    ]
    expected_exception = Exception("Test Error")
    mock_popen.side_effect = [expected_exception, None]
    with _LocalJobContainer(image_uri, aws_session, mock_logger) as container:
        container.run_local_job(env_variables)
        assert container.run_log == expected_exception
    assert mock_check_output.call_count == 3
    mock_run.assert_called_with(["docker", "stop", running_container_name])
    mock_logger.error.assert_called_with(expected_exception)


@patch("subprocess.check_output")
@patch("subprocess.run")
def test_make_dir(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    test_dir_path = "/test/dir/path"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(""),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.makedir(test_dir_path)
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    mock_check_output.assert_any_call(
        ["docker", "exec", running_container_name, "mkdir", "-p", test_dir_path]
    )
    assert mock_check_output.call_count == 3
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 1


@patch("subprocess.check_output")
@patch("subprocess.run")
def test_copy_to(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    source_path = str(Path("test", "source", "dir", "path", "srcfile.txt"))
    dest_path = str(PurePosixPath("test", "dest", "dir", "path", "dstfile.txt"))
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(""),
        str.encode(""),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.copy_to(source_path, dest_path)
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    mock_check_output.assert_any_call(
        [
            "docker",
            "exec",
            running_container_name,
            "mkdir",
            "-p",
            str(PurePosixPath("test", "dest", "dir", "path")),
        ]
    )
    mock_check_output.assert_any_call(
        ["docker", "cp", source_path, f"{running_container_name}:{dest_path}"]
    )
    assert mock_check_output.call_count == 4
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 1


@patch("subprocess.check_output")
@patch("subprocess.run")
def test_copy_from(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    source_path = "/test/source/dir/path/srcfile.txt"
    dest_path = "/test/dest/dir/path/dstfile.txt"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(""),
        str.encode(""),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.copy_from(source_path, dest_path)
    mock_check_output.assert_any_call(["docker", "images", "-q", image_uri])
    mock_check_output.assert_any_call(
        ["docker", "run", "-d", "--rm", local_image_name, "tail", "-f", "/dev/null"]
    )
    mock_check_output.assert_any_call(
        ["docker", "cp", f"{running_container_name}:{source_path}", dest_path]
    )
    assert mock_check_output.call_count == 3
    mock_run.assert_any_call(["docker", "stop", running_container_name])
    assert mock_run.call_count == 1


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=ValueError)
def test_run_fails_no_program(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    env_variables = {
        "ENV0": "VALUE0",
        "ENV1": "VALUE1",
    }
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        str.encode(""),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.run_local_job(env_variables)


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=subprocess.CalledProcessError)
def test_make_dir_fails(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    test_dir_path = "/test/dir/path"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        subprocess.CalledProcessError("Test Error", "test", str.encode("test output")),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.makedir(test_dir_path)


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=subprocess.CalledProcessError)
def test_copy_to_fails(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    source_path = "/test/source/dir/path/srcfile.txt"
    dest_path = "/test/dest/dir/path/dstfile.txt"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        subprocess.CalledProcessError("Test Error", "test", str.encode("test output")),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.copy_to(source_path, dest_path)


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=subprocess.CalledProcessError)
def test_copy_from_fails(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    source_path = "/test/source/dir/path/srcfile.txt"
    dest_path = "/test/dest/dir/path/dstfile.txt"
    mock_check_output.side_effect = [
        str.encode(local_image_name),
        str.encode(running_container_name),
        subprocess.CalledProcessError("Test Error", "test", str.encode("test output")),
    ]
    with _LocalJobContainer(image_uri, aws_session) as container:
        container.copy_from(source_path, dest_path)


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=ValueError)
def test_pull_fails_no_auth(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    mock_check_output.side_effect = [
        str.encode(""),
        str.encode(local_image_name),
        str.encode(running_container_name),
    ]
    aws_session.ecr_client.get_authorization_token.return_value = {}
    with _LocalJobContainer(image_uri, aws_session):
        pass


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=ValueError)
def test_pull_fails_invalid_uri(mock_run, mock_check_output, aws_session):
    local_image_name = "LocalImageName"
    running_container_name = "RunningContainer"
    mock_check_output.side_effect = [
        str.encode(""),
        str.encode(local_image_name),
        str.encode(running_container_name),
    ]
    aws_session.ecr_client.get_authorization_token.return_value = {}
    with _LocalJobContainer("TestURI", aws_session):
        pass


@patch("subprocess.check_output")
@patch("subprocess.run")
@pytest.mark.xfail(raises=ValueError)
def test_pull_fails_unknown_reason(mock_run, mock_check_output, repo_uri, image_uri, aws_session):
    test_token = "Test Token"
    mock_check_output.side_effect = [
        str.encode(""),
        str.encode(""),
    ]
    aws_session.ecr_client.get_authorization_token.return_value = {
        "authorizationData": [{"authorizationToken": base64.b64encode(str.encode(test_token))}]
    }
    with _LocalJobContainer(image_uri, aws_session):
        pass



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable

import numpy as np
from scipy.linalg import fractional_matrix_power

from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.default_simulator.linalg_utils import multiply_matrix
from braket.registers.qubit_set import QubitSet


def calculate_unitary_big_endian(
    instructions: Iterable[Instruction], qubits: QubitSet
) -> np.ndarray:
    """Returns the unitary matrix representation for all the `instruction`s on qubits `qubits`.

    Note:
        The performance of this method degrades with qubit count. It might be slow for
        `len(qubits)` > 10.

    Args:
        instructions (Iterable[Instruction]): The instructions for which the unitary matrix
            will be calculated.
        qubits (QubitSet): The actual qubits used by the instructions.

    Returns:
        np.ndarray: A numpy array with shape (2^qubit_count, 2^qubit_count) representing the
        `instructions` as a unitary matrix.

    Raises:
        TypeError: If `instructions` is not composed only of `Gate` instances,
            i.e. a circuit with `Noise` operators will raise this error.
            Any `CompilerDirective` instructions will be ignored, as these should
            not affect the unitary representation of the circuit.
    """
    qubits_sorted = sorted(qubits)
    qubit_count = len(qubits_sorted)
    # Map qubits to contiguous indices starting from 0
    index_substitutions = {qubits_sorted[i]: i for i in range(qubit_count)}
    rank = 2**qubit_count
    # Initialize identity unitary as type (rank, rank) tensor
    unitary = np.eye(rank, dtype=complex).reshape([2] * 2 * qubit_count)

    for instruction in instructions:
        if isinstance(instruction.operator, CompilerDirective):
            continue
        if not isinstance(instruction.operator, Gate):
            raise TypeError("Only Gate operators are supported to build the unitary")

        base_gate_matrix = instruction.operator.to_matrix()
        if int(instruction.power) == instruction.power:
            gate_matrix = np.linalg.matrix_power(base_gate_matrix, int(instruction.power))
        else:
            gate_matrix = fractional_matrix_power(base_gate_matrix, instruction.power)

        unitary = multiply_matrix(
            unitary,
            gate_matrix,
            tuple(index_substitutions[qubit] for qubit in instruction.target),
            controls=instruction.control,
            control_state=instruction.control_state,
        )

    return unitary.reshape(rank, rank)



run_entry_point = """
import cloudpickle
import os
from braket.jobs import get_results_dir, save_job_result
from braket.jobs_data import PersistedJobDataFormat


# load and run serialized entry point function
recovered = cloudpickle.loads({serialized})
def {function_name}():
    try:
        # set working directory to results dir
        os.chdir(get_results_dir())

        # create symlinks to input data
        links = link_input()

        result = recovered()
    finally:
        clean_links(links)
    if result is not None:
        save_job_result(result, data_format=PersistedJobDataFormat.PICKLED_V4)
    return result
"""

symlink_input_data = '''
from pathlib import Path
from braket.jobs import get_input_data_dir


def make_link(input_link_path, input_data_path, links):
    """ Create symlink from input_link_path to input_data_path. """
    input_link_path.parent.mkdir(parents=True, exist_ok=True)
    input_link_path.symlink_to(input_data_path)
    print(input_link_path, '->', input_data_path)
    links[input_link_path] = input_data_path


def link_input():
    links = {{}}
    dirs = set()
    # map of data sources to lists of matched local files
    prefix_matches = {prefix_matches}

    for channel, data in {input_data_items}:

        if channel in {prefix_channels}:
            # link all matched files
            for input_link_name in prefix_matches[channel]:
                input_link_path = Path(input_link_name)
                input_data_path = Path(get_input_data_dir(channel)) / input_link_path.name
                make_link(input_link_path, input_data_path, links)

        else:
            input_link_path = Path(data)
            if channel in {directory_channels}:
                # link directory source directly to input channel directory
                input_data_path = Path(get_input_data_dir(channel))
            else:
                # link file source to file within input channel directory
                input_data_path = Path(get_input_data_dir(channel), Path(data).name)
            make_link(input_link_path, input_data_path, links)

    return links


def clean_links(links):
    for link, target in links.items():
        if link.is_symlink and link.readlink() == target:
            link.unlink()

        if link.is_relative_to(Path()):
            for dir in link.parents[:-1]:
                try:
                    dir.rmdir()
                except:
                    # directory not empty
                    pass
'''



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np

from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.circuits.moments import Moments
from braket.circuits.noise import Noise
from braket.circuits.quantum_operator_helpers import is_unitary
from braket.registers.qubit_set import QubitSet, QubitSetInput

if TYPE_CHECKING:  # pragma: no cover
    from braket.circuits.circuit import Circuit


def no_noise_applied_warning(noise_applied: bool) -> None:
    """Helper function to give a warning is noise is not applied.

    Args:
        noise_applied (bool): True if the noise has been applied.
    """
    if noise_applied is False:
        warnings.warn(
            "Noise is not applied to any gate, as there is no eligible gate in the circuit"
            " with the input criteria or there is no multi-qubit gate to apply"
            " the multi-qubit noise.",
            stacklevel=1,
        )


def wrap_with_list(an_item: Any) -> list[Any]:
    """Helper function to make the input parameter a list.

    Args:
        an_item (Any): The item to wrap.

    Returns:
        list[Any]: The item wrapped in a list.
    """
    if an_item is not None and not isinstance(an_item, list):
        an_item = [an_item]
    return an_item


def check_noise_target_gates(noise: Noise, target_gates: Iterable[type[Gate]]) -> None:
    """Helper function to check
    1. whether all the elements in target_gates are a Gate type;
    2. if `noise` is multi-qubit noise and `target_gates` contain gates
    with the number of qubits is the same as `noise.qubit_count`.

    Args:
        noise (Noise): A Noise class object to be applied to the circuit.
        target_gates (Iterable[type[Gate]]): Gate class or
            List of Gate classes which `noise` is applied to.
    """
    if not all(isinstance(g, type) and issubclass(g, Gate) for g in target_gates):
        raise TypeError("All elements in target_gates must be an instance of the Gate class")

    if noise.qubit_count > 1:
        for g in target_gates:
            fixed_qubit_count = g.fixed_qubit_count()
            if fixed_qubit_count is NotImplemented:
                raise ValueError(
                    f"Target gate {g} can be instantiated on a variable number of qubits,"
                    " but noise can only target gates with fixed qubit counts."
                )
            if fixed_qubit_count != noise.qubit_count:
                raise ValueError(
                    f"Target gate {g} acts on {fixed_qubit_count} qubits,"
                    f" but {noise} acts on {noise.qubit_count} qubits."
                )


def check_noise_target_unitary(noise: Noise, target_unitary: np.ndarray) -> None:
    """Helper function to check
    1. whether the input matrix is a np.ndarray type;
    2. whether the target_unitary is a unitary;

    Args:
        noise (Noise): A Noise class object to be applied to the circuit.
        target_unitary (ndarray): matrix of the target unitary gates
    """
    if not isinstance(target_unitary, np.ndarray):
        raise TypeError("target_unitary must be a np.ndarray type")

    if not is_unitary(target_unitary):
        raise ValueError("target_unitary must be a unitary")


def check_noise_target_qubits(
    circuit: Circuit, target_qubits: Optional[QubitSetInput] = None
) -> QubitSet:
    """Helper function to check whether all the target_qubits are positive integers.

    Args:
        circuit (Circuit): A ciruit where `noise` is to be checked.
        target_qubits (Optional[QubitSetInput]): Index or indices of qubit(s).

    Returns:
        QubitSet: The target qubits.
    """
    if target_qubits is None:
        target_qubits = circuit.qubits
    else:
        target_qubits = wrap_with_list(target_qubits)
        if not all(isinstance(q, int) for q in target_qubits):
            raise TypeError("target_qubits must be integer(s)")
        if not all(q >= 0 for q in target_qubits):
            raise ValueError("target_qubits must contain only non-negative integers.")

        target_qubits = QubitSet(target_qubits)

    return target_qubits


def apply_noise_to_moments(
    circuit: Circuit, noise: Iterable[type[Noise]], target_qubits: QubitSet, position: str
) -> Circuit:
    """Apply initialization/readout noise to the circuit.

    When `noise.qubit_count` == 1, `noise` is added to all qubits in `target_qubits`.

    When `noise.qubit_count` > 1, `noise.qubit_count` must be the same as the length of
    `target_qubits`.

    Args:
        circuit (Circuit): A ciruit where `noise` is applied to.
        noise (Iterable[type[Noise]]): Noise channel(s) to be applied
            to the circuit.
        target_qubits (QubitSet): Index or indices of qubits. `noise` is applied to.
        position (str): The position to add the noise to. May be 'initialization' or
            'readout_noise'.

    Returns:
        Circuit: modified circuit.
    """
    noise_instructions = []
    for noise_channel in noise:
        if noise_channel.qubit_count == 1:
            new = [Instruction(noise_channel, qubit) for qubit in target_qubits]
            noise_instructions = noise_instructions + new
        else:
            noise_instructions.append(Instruction(noise_channel, target_qubits))

    new_moments = Moments()

    if position == "initialization":
        for noise in noise_instructions:
            new_moments.add_noise(noise, "initialization_noise")

    # add existing instructions
    for moment_key in circuit.moments:
        instruction = circuit.moments[moment_key]
        # if the instruction is noise instruction
        if isinstance(instruction.operator, Noise):
            new_moments.add_noise(instruction, moment_key.moment_type, moment_key.noise_index)
        # if the instruction is a gate instruction
        else:
            new_moments.add([instruction], moment_key.noise_index)

    if position == "readout":
        for noise in noise_instructions:
            new_moments.add_noise(noise, "readout_noise")

    circuit._moments = new_moments

    return circuit


def _apply_noise_to_gates_helper(
    noise: Iterable[type[Noise]],
    target_qubits: QubitSet,
    instruction: Instruction,
    noise_index: int,
    intersection: QubitSet,
    noise_applied: bool,
    new_noise_instruction: Iterable,
) -> tuple[Iterable[Instruction], int, bool]:
    """Helper function to work out the noise instructions to be attached to a gate.

    Args:
        noise (Iterable[type[Noise]]): Noise channel(s) to be applied
            to the circuit.
        target_qubits (QubitSet): Index or indices of qubits which `noise` is applied to.
        instruction (Instruction): Instruction of the gate which `noise` is applied to.
        noise_index (int): The number of noise channels applied to the gate.
        intersection (QubitSet): Intersection of target_qubits and the qubits associated
            with the gate.
        noise_applied (bool): Whether noise is applied or not.
        new_noise_instruction (Iterable): current new noise instructions to be attached
            to the circuit.

    Returns:
        tuple[Iterable[Instruction], int, bool]: A tuple of three values:
        new_noise_instruction: A list of noise intructions
        noise_index: The number of noise channels applied to the gate
        noise_applied: Whether noise is applied or not
    """
    for noise_channel in noise:
        if noise_channel.qubit_count == 1:
            for qubit in intersection:
                # apply noise to the qubit if it is in target_qubits
                noise_index += 1
                new_noise_instruction.append((Instruction(noise_channel, qubit), noise_index))
                noise_applied = True
        # only apply noise to the gates that have the same qubit_count as the noise.
        elif (
            instruction.operator.qubit_count == noise_channel.qubit_count
            and instruction.target.issubset(target_qubits)
        ):
            noise_index += 1
            new_noise_instruction.append(
                (Instruction(noise_channel, instruction.target), noise_index)
            )
            noise_applied = True

    return new_noise_instruction, noise_index, noise_applied


def apply_noise_to_gates(
    circuit: Circuit,
    noise: Iterable[type[Noise]],
    target_gates: Union[Iterable[type[Gate]], np.ndarray],
    target_qubits: QubitSet,
) -> Circuit:
    """Apply noise after target gates in target qubits.

    When `noise.qubit_count` == 1, `noise` is applied to target_qubits after `target_gates`.

    When `noise.qubit_count` > 1, all elements in `target_gates`, if is given, must have
    the same number of qubits as `noise.qubit_count`.

    Args:
        circuit (Circuit): A ciruit where `noise` is applied to.
        noise (Iterable[type[Noise]]): Noise channel(s) to be applied
            to the circuit.
        target_gates (Union[Iterable[type[Gate]], ndarray]): List of gates, or a unitary matrix
            which `noise` is applied to.
        target_qubits (QubitSet): Index or indices of qubits which `noise` is applied to.

    Returns:
        Circuit: modified circuit.

    Raises:
        Warning:
            If `noise` is multi-qubit noise while there is no gate with the same
            number of qubits in `target_qubits` or in the whole circuit when
            `target_qubits` is not given.
            If no `target_gates` exist in `target_qubits` or in the whole circuit
            when `target_qubits` is not given.
    """
    new_moments = Moments()
    noise_applied = False

    for moment_key in circuit.moments:
        instruction = circuit.moments[moment_key]

        # add the instruction to new_moments if it is noise instruction
        if isinstance(instruction.operator, Noise):
            new_moments.add_noise(instruction, moment_key.moment_type, moment_key.noise_index)

        # if the instruction is a gate instruction
        else:
            new_noise_instruction = []
            noise_index = moment_key.noise_index
            if isinstance(target_gates, np.ndarray):
                if (instruction.operator.name == "Unitary") and (
                    np.array_equiv(instruction.operator._matrix, target_gates)
                ):
                    intersection = list(set(instruction.target) & set(target_qubits))
                    (
                        new_noise_instruction,
                        noise_index,
                        noise_applied,
                    ) = _apply_noise_to_gates_helper(
                        noise,
                        target_qubits,
                        instruction,
                        noise_index,
                        intersection,
                        noise_applied,
                        new_noise_instruction,
                    )

            elif (target_gates is None) or (
                instruction.operator.name in [g.__name__ for g in target_gates]
            ):
                intersection = list(set(instruction.target) & set(target_qubits))
                new_noise_instruction, noise_index, noise_applied = _apply_noise_to_gates_helper(
                    noise,
                    target_qubits,
                    instruction,
                    noise_index,
                    intersection,
                    noise_applied,
                    new_noise_instruction,
                )

            # add the gate and gate noise instructions to new_moments
            new_moments.add([instruction], noise_index=noise_index)
            for instruction, noise_index in new_noise_instruction:
                new_moments.add_noise(instruction, "gate_noise", noise_index)

    no_noise_applied_warning(noise_applied)
    circuit._moments = new_moments
    return circuit



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import os.path
import re
import sys
import tempfile
import time
from pathlib import Path

import job_test_script
import pytest
from job_test_module.job_test_submodule.job_test_submodule_file import submodule_helper

from braket.aws import AwsSession
from braket.aws.aws_quantum_job import AwsQuantumJob
from braket.devices import Devices
from braket.jobs import Framework, get_input_data_dir, hybrid_job, retrieve_image, save_job_result


def decorator_python_version():
    aws_session = AwsSession()
    image_uri = retrieve_image(Framework.BASE, aws_session.region)
    tag = aws_session.get_full_image_tag(image_uri)
    major_version, minor_version = re.search(r"-py(\d)(\d+)-", tag).groups()
    return int(major_version), int(minor_version)


def test_failed_quantum_job(aws_session, capsys, failed_quantum_job):
    """Asserts the hybrid job is failed with the output, checkpoints,
    quantum tasks not created in bucket and only input is uploaded to s3. Validate the
    results/download results have the response raising RuntimeError. Also,
    check if the logs displays the Assertion Error.
    """
    job = failed_quantum_job
    job_name = job.name

    pattern = f"^arn:aws:braket:{aws_session.region}:\\d{{12}}:job/[a-z0-9-]+$"
    assert re.match(pattern=pattern, string=job.arn)

    # Check job is in failed state.
    while True:
        time.sleep(5)
        if job.state() in AwsQuantumJob.TERMINAL_STATES:
            break
    assert job.state(use_cached_value=True) == "FAILED"

    # Check whether the respective folder with files are created for script,
    # output, tasks and checkpoints.
    job_name = job.name
    s3_bucket = aws_session.default_bucket()
    subdirectory = re.match(
        rf"s3://{s3_bucket}/jobs/{job.name}/(\d+)/script/source.tar.gz",
        job.metadata()["algorithmSpecification"]["scriptModeConfig"]["s3Uri"],
    ).group(1)
    keys = aws_session.list_keys(
        bucket=s3_bucket,
        prefix=f"jobs/{job_name}/{subdirectory}/",
    )
    assert keys == [f"jobs/{job_name}/{subdirectory}/script/source.tar.gz"]

    # no results saved
    assert job.result() == {}

    job.logs()
    log_data, errors = capsys.readouterr()
    assert errors == ""
    logs_to_validate = [
        "Invoking script with the following command:",
        "braket_container.py",
        "Running Code As Process",
        "Test job started!!!!!",
        "FileNotFoundError: [Errno 2] No such file or directory: 'fake_file'",
        "Code Run Finished",
        '"user_entry_point": "braket_container.py"',
    ]

    for data in logs_to_validate:
        assert data in log_data

    assert job.metadata()["failureReason"] == (
        "AlgorithmError: FileNotFoundError: [Errno 2] "
        "No such file or directory: 'fake_file', exit code: 1"
    )


def test_completed_quantum_job(aws_session, capsys, completed_quantum_job):
    """Asserts the hybrid job is completed with the output, checkpoints, quantum tasks and
    script folder created in S3 for respective hybrid job. Validate the results are
    downloaded and results are what we expect. Also, assert that logs contains all the
    necessary steps for setup and running the hybrid job and is displayed to the user.
    """

    job = completed_quantum_job
    job_name = job.name
    pattern = f"^arn:aws:braket:{aws_session.region}:\\d{{12}}:job/[a-z0-9-]+$"
    assert re.match(pattern=pattern, string=job.arn)

    # Check the job has completed
    job.result()

    assert job.state(use_cached_value=True) == "COMPLETED"

    # Check whether the respective folder with files are created for script,
    # output, tasks and checkpoints.
    job_name = job.name
    s3_bucket = aws_session.default_bucket()
    subdirectory = re.match(
        rf"s3://{s3_bucket}/jobs/{job.name}/(\d+)/script/source.tar.gz",
        job.metadata()["algorithmSpecification"]["scriptModeConfig"]["s3Uri"],
    ).group(1)
    keys = aws_session.list_keys(
        bucket=s3_bucket,
        prefix=f"jobs/{job_name}/{subdirectory}/",
    )
    for expected_key in [
        f"jobs/{job_name}/{subdirectory}/script/source.tar.gz",
        f"jobs/{job_name}/{subdirectory}/data/output/model.tar.gz",
        f"jobs/{job_name}/{subdirectory}/checkpoints/{job_name}_plain_data.json",
        f"jobs/{job_name}/{subdirectory}/checkpoints/{job_name}.json",
    ]:
        assert any(re.match(expected_key, key) for key in keys)

    # Check that tasks exist in the correct location
    tasks_keys = aws_session.list_keys(
        bucket=s3_bucket,
        prefix=f"jobs/{job_name}/tasks/",
    )
    expected_task_location = f"jobs/{job_name}/tasks/[^/]*/results.json"
    assert any(re.match(expected_task_location, key) for key in tasks_keys)

    # Check if checkpoint is uploaded in requested format.
    for s3_key, expected_data in [
        (
            f"jobs/{job_name}/{subdirectory}/checkpoints/{job_name}_plain_data.json",
            {
                "braketSchemaHeader": {
                    "name": "braket.jobs_data.persisted_job_data",
                    "version": "1",
                },
                "dataDictionary": {"some_data": "abc"},
                "dataFormat": "plaintext",
            },
        ),
        (
            f"jobs/{job_name}/{subdirectory}/checkpoints/{job_name}.json",
            {
                "braketSchemaHeader": {
                    "name": "braket.jobs_data.persisted_job_data",
                    "version": "1",
                },
                "dataDictionary": {"some_data": "gASVBwAAAAAAAACMA2FiY5Qu\n"},
                "dataFormat": "pickled_v4",
            },
        ),
    ]:
        assert (
            json.loads(
                aws_session.retrieve_s3_object_body(s3_bucket=s3_bucket, s3_object_key=s3_key)
            )
            == expected_data
        )

    current_dir = Path.cwd()

    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)
        try:
            # Check results match the expectations.
            assert job.result() == {"converged": True, "energy": -0.2}
        finally:
            os.chdir(current_dir)

    # Check the logs and validate it contains required output.
    job.logs(wait=True)
    log_data, errors = capsys.readouterr()
    assert errors == ""
    logs_to_validate = [
        "Invoking script with the following command:",
        "braket_container.py",
        "Running Code As Process",
        "Test job started!!!!!",
        "Test job completed!!!!!",
        "Code Run Finished",
        '"user_entry_point": "braket_container.py"',
        "Reporting training SUCCESS",
    ]

    for data in logs_to_validate:
        assert data in log_data


@pytest.mark.xfail(
    (sys.version_info.major, sys.version_info.minor) != decorator_python_version(),
    raises=RuntimeError,
    reason="Python version mismatch",
)
def test_decorator_job():
    class MyClass:
        attribute = "value"

        def __str__(self):
            return f"MyClass({self.attribute})"

    @hybrid_job(
        device=Devices.Amazon.SV1,
        include_modules="job_test_script",
        dependencies=str(Path("test", "integ_tests", "requirements.txt")),
        input_data=str(Path("test", "integ_tests", "requirements")),
    )
    def decorator_job(a, b: int, c=0, d: float = 1.0, **extras):
        with open(Path(get_input_data_dir()) / "requirements.txt", "r") as f:
            assert f.readlines() == ["pytest\n"]
        with open(Path("test", "integ_tests", "requirements.txt"), "r") as f:
            assert f.readlines() == ["pytest\n"]
        assert dir(pytest)
        assert a.attribute == "value"
        assert b == 2
        assert c == 0
        assert d == 5
        assert extras["extra_arg"] == "extra_value"

        hp_file = os.environ["AMZN_BRAKET_HP_FILE"]
        with open(hp_file, "r") as f:
            hyperparameters = json.load(f)
        assert hyperparameters == {
            "a": "MyClass{value}",
            "b": "2",
            "c": "0",
            "d": "5",
            "extra_arg": "extra_value",
        }

        with open("test/output_file.txt", "w") as f:
            f.write("hello")

        return job_test_script.job_helper()

    job = decorator_job(MyClass(), 2, d=5, extra_arg="extra_value")
    assert job.result()["status"] == "SUCCESS"

    current_dir = Path.cwd()
    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)
        try:
            job.download_result()
            with open(Path(job.name, "test", "output_file.txt"), "r") as f:
                assert f.read() == "hello"
            assert (
                Path(job.name, "results.json").exists()
                and Path(job.name, "test").exists()
                and not Path(job.name, "test", "integ_tests").exists()
            )
        finally:
            os.chdir(current_dir)


@pytest.mark.xfail(
    (sys.version_info.major, sys.version_info.minor) != decorator_python_version(),
    raises=RuntimeError,
    reason="Python version mismatch",
)
def test_decorator_job_submodule():
    @hybrid_job(
        device=Devices.Amazon.SV1,
        include_modules=[
            "job_test_module",
        ],
        dependencies=Path(
            "test", "integ_tests", "job_test_module", "job_test_submodule", "requirements.txt"
        ),
        input_data={
            "my_input": str(Path("test", "integ_tests", "requirements.txt")),
            "my_dir": str(Path("test", "integ_tests", "job_test_module")),
        },
    )
    def decorator_job_submodule():
        with open(Path(get_input_data_dir("my_input")) / "requirements.txt", "r") as f:
            assert f.readlines() == ["pytest\n"]
        with open(Path("test", "integ_tests", "requirements.txt"), "r") as f:
            assert f.readlines() == ["pytest\n"]
        with open(
            Path(get_input_data_dir("my_dir")) / "job_test_submodule" / "requirements.txt", "r"
        ) as f:
            assert f.readlines() == ["pytest\n"]
        with open(
            Path(
                "test",
                "integ_tests",
                "job_test_module",
                "job_test_submodule",
                "requirements.txt",
            ),
            "r",
        ) as f:
            assert f.readlines() == ["pytest\n"]
        assert dir(pytest)
        save_job_result(submodule_helper())

    job = decorator_job_submodule()
    assert job.result()["status"] == "SUCCESS"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.jobs.config import (  # noqa: F401
    CheckpointConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.data_persistence import (  # noqa: F401
    load_job_checkpoint,
    load_job_result,
    save_job_checkpoint,
    save_job_result,
)
from braket.jobs.environment_variables import (  # noqa: F401
    get_checkpoint_dir,
    get_hyperparameters,
    get_input_data_dir,
    get_job_device_arn,
    get_job_name,
    get_results_dir,
)
from braket.jobs.hybrid_job import hybrid_job  # noqa: F401
from braket.jobs.image_uris import Framework, retrieve_image  # noqa: F401



import ast
import importlib
import re
import sys
import tempfile
from logging import getLogger
from pathlib import Path
from ssl import PROTOCOL_TLS_CLIENT, SSLContext
from unittest.mock import MagicMock, patch

import job_module
import pytest
from cloudpickle import cloudpickle

from braket.aws import AwsQuantumJob
from braket.devices import Devices
from braket.jobs import hybrid_job
from braket.jobs.config import (
    CheckpointConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.hybrid_job import _sanitize, _serialize_entry_point
from braket.jobs.local import LocalQuantumJob


@pytest.fixture
def aws_session():
    aws_session = MagicMock()
    python_version_str = f"py{sys.version_info.major}{sys.version_info.minor}"
    aws_session.get_full_image_tag.return_value = f"1.0-cpu-{python_version_str}-ubuntu22.04"
    aws_session.region = "us-west-2"
    return aws_session


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_decorator_defaults(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(device=None, aws_session=aws_session)
    def my_entry(c=0, d: float = 1.0, **extras):
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
    wait_until_complete = False

    device = "local:none/none"

    my_entry()

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        hyperparameters={"c": "0", "d": "1.0"},
        logger=getLogger("braket.jobs.hybrid_job"),
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


@pytest.mark.parametrize("include_modules", (job_module, ["job_module"]))
@patch("braket.jobs.image_uris.retrieve_image")
@patch("sys.stdout")
@patch("time.time", return_value=123.0)
@patch("cloudpickle.register_pickle_by_value")
@patch("cloudpickle.unregister_pickle_by_value")
@patch("shutil.copy")
@patch("builtins.open")
@patch.object(AwsQuantumJob, "create")
def test_decorator_non_defaults(
    mock_create,
    _mock_open,
    mock_copy,
    mock_register,
    mock_unregister,
    mock_time,
    mock_stdout,
    mock_retrieve,
    include_modules,
):
    mock_retrieve.return_value = "should-not-be-used"
    dependencies = "my_requirements.txt"
    image_uri = "my_image.uri"
    default_instance = InstanceConfig()
    distribution = "data_parallel"
    copy_checkpoints_from_job = "arn/other-job"
    checkpoint_config = CheckpointConfig(localPath="local", s3Uri="s3")
    role_arn = "role_arn"
    stopping_condition = StoppingCondition(maxRuntimeInSeconds=10)
    output_data_config = OutputDataConfig(s3Path="s3")
    aws_session = MagicMock()
    tags = {"my_tag": "my_value"}
    reservation_arn = (
        "arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
    )
    logger = getLogger(__name__)

    with tempfile.TemporaryDirectory() as tempdir:
        Path(tempdir, "temp_dir").mkdir()
        Path(tempdir, "temp_file").touch()

        input_data = {
            "my_prefix": "my_input_data",
            "my_dir": Path(tempdir, "temp_dir"),
            "my_file": Path(tempdir, "temp_file"),
            "my_s3_prefix": "s3://bucket/path/to/prefix",
            "my_s3_config": S3DataSourceConfig(s3_data="s3://bucket/path/to/prefix"),
        }

        @hybrid_job(
            device=Devices.Amazon.SV1,
            include_modules=include_modules,
            dependencies=dependencies,
            image_uri=image_uri,
            input_data=input_data,
            wait_until_complete=True,
            instance_config=default_instance,
            distribution=distribution,
            checkpoint_config=checkpoint_config,
            copy_checkpoints_from_job=copy_checkpoints_from_job,
            role_arn=role_arn,
            stopping_condition=stopping_condition,
            output_data_config=output_data_config,
            aws_session=aws_session,
            tags=tags,
            reservation_arn=reservation_arn,
            logger=logger,
        )
        def my_entry(a, b: int, c=0, d: float = 1.0, **extras) -> str:
            return "my entry return value"

        mock_tempdir = MagicMock(spec=tempfile.TemporaryDirectory)
        mock_tempdir_name = "job_temp_dir_00000"
        mock_tempdir.__enter__.return_value = mock_tempdir_name

        device = Devices.Amazon.SV1
        source_module = mock_tempdir_name
        entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
        wait_until_complete = True

        s3_not_linked = (
            "Input data channels mapped to an S3 source will not be available in the working "
            'directory. Use `get_input_data_dir(channel="my_s3_prefix")` to read input data '
            "from S3 source inside the job container."
        )

        with patch("tempfile.TemporaryDirectory", return_value=mock_tempdir):
            my_entry("a", 2, 3, 4, extra_param="value", another=6)

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        image_uri=image_uri,
        input_data=input_data,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        instance_config=default_instance,
        distribution=distribution,
        hyperparameters={
            "a": "a",
            "b": "2",
            "c": "3",
            "d": "4",
            "extra_param": "value",
            "another": "6",
        },
        checkpoint_config=checkpoint_config,
        copy_checkpoints_from_job=copy_checkpoints_from_job,
        role_arn=role_arn,
        stopping_condition=stopping_condition,
        output_data_config=output_data_config,
        aws_session=aws_session,
        tags=tags,
        logger=logger,
        reservation_arn=reservation_arn,
    )
    included_module = importlib.import_module("job_module")
    mock_register.assert_called_with(included_module)
    mock_unregister.assert_called_with(included_module)
    mock_copy.assert_called_with(
        Path("my_requirements.txt").resolve(), Path(mock_tempdir_name, "requirements.txt")
    )
    assert mock_tempdir.__exit__.called
    mock_stdout.write.assert_any_call(s3_not_linked)


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_decorator_non_dict_input(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"
    input_prefix = "my_input"

    @hybrid_job(device=None, input_data=input_prefix, aws_session=aws_session)
    def my_entry():
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
    wait_until_complete = False

    device = "local:none/none"

    my_entry()

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        hyperparameters={},
        logger=getLogger("braket.jobs.hybrid_job"),
        input_data=input_prefix,
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_decorator_list_dependencies(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"
    dependency_list = ["dep_1", "dep_2", "dep_3"]

    @hybrid_job(
        device=None,
        aws_session=aws_session,
        dependencies=dependency_list,
    )
    def my_entry(c=0, d: float = 1.0, **extras):
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
    wait_until_complete = False

    device = "local:none/none"

    my_entry()

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        hyperparameters={"c": "0", "d": "1.0"},
        logger=getLogger("braket.jobs.hybrid_job"),
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called
    _mock_open.assert_called_with(Path(mock_tempdir_name) / "requirements.txt", "w")
    _mock_open.return_value.__enter__.return_value.write.assert_called_with(
        "\n".join(dependency_list)
    )


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(LocalQuantumJob, "create")
def test_decorator_local(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(device=Devices.Amazon.SV1, local=True, aws_session=aws_session)
    def my_entry():
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    device = Devices.Amazon.SV1
    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"

    my_entry()

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        job_name="my-entry-123000",
        hyperparameters={},
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(LocalQuantumJob, "create")
def test_decorator_local_unsupported_args(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(
        device=Devices.Amazon.SV1,
        local=True,
        wait_until_complete=True,
        copy_checkpoints_from_job="arn/other-job",
        instance_config=InstanceConfig(),
        distribution="data_parallel",
        stopping_condition=StoppingCondition(),
        tags={"my_tag": "my_value"},
        logger=getLogger(__name__),
        aws_session=aws_session,
    )
    def my_entry():
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    device = Devices.Amazon.SV1
    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"

    my_entry()

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        job_name="my-entry-123000",
        hyperparameters={},
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_job_name_too_long(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(device="local:braket/default", aws_session=aws_session)
    def this_is_a_50_character_func_name_for_testing_names():
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    device = "local:braket/default"
    source_module = mock_tempdir_name
    entry_point = (
        f"{mock_tempdir_name}.entry_point:this_is_a_50_character_func_name_for_testing_names"
    )
    wait_until_complete = False

    with pytest.warns(UserWarning):
        this_is_a_50_character_func_name_for_testing_names()

        expected_job_name = "this-is-a-50-character-func-name-for-testin-123000"

        mock_create.assert_called_with(
            device=device,
            source_module=source_module,
            entry_point=entry_point,
            wait_until_complete=wait_until_complete,
            job_name=expected_job_name,
            hyperparameters={},
            logger=getLogger("braket.jobs.hybrid_job"),
            aws_session=aws_session,
        )
        assert len(expected_job_name) == 50
        assert mock_tempdir.return_value.__exit__.called


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_decorator_pos_only_slash(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(device="local:braket/default", aws_session=aws_session)
    def my_entry(pos_only, /):
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    device = "local:braket/default"
    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
    wait_until_complete = False

    pos_only_warning = "Positional only arguments will not be logged to the hyperparameters file."
    with pytest.warns(match=pos_only_warning):
        my_entry("pos_only")

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        hyperparameters={},
        logger=getLogger("braket.jobs.hybrid_job"),
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


@patch.object(sys.modules["braket.jobs.hybrid_job"], "retrieve_image")
@patch("time.time", return_value=123.0)
@patch("builtins.open")
@patch("tempfile.TemporaryDirectory")
@patch.object(AwsQuantumJob, "create")
def test_decorator_pos_only_args(
    mock_create, mock_tempdir, _mock_open, mock_time, mock_retrieve, aws_session
):
    mock_retrieve.return_value = "00000000.dkr.ecr.us-west-2.amazonaws.com/latest"

    @hybrid_job(device="local:braket/default", aws_session=aws_session)
    def my_entry(*args):
        return "my entry return value"

    mock_tempdir_name = "job_temp_dir_00000"
    mock_tempdir.return_value.__enter__.return_value = mock_tempdir_name

    device = "local:braket/default"
    source_module = mock_tempdir_name
    entry_point = f"{mock_tempdir_name}.entry_point:my_entry"
    wait_until_complete = False

    pos_only_warning = "Positional only arguments will not be logged to the hyperparameters file."
    with pytest.warns(match=pos_only_warning):
        my_entry("pos_only")

    mock_create.assert_called_with(
        device=device,
        source_module=source_module,
        entry_point=entry_point,
        wait_until_complete=wait_until_complete,
        job_name="my-entry-123000",
        hyperparameters={},
        logger=getLogger("braket.jobs.hybrid_job"),
        aws_session=aws_session,
    )
    assert mock_tempdir.return_value.__exit__.called


def test_serialization_error(aws_session):
    ssl_context = SSLContext(protocol=PROTOCOL_TLS_CLIENT)

    @hybrid_job(device=None, aws_session=aws_session)
    def fails_serialization():
        print(ssl_context)

    serialization_failed = (
        "Serialization failed for decorator hybrid job. If you are referencing "
        "an object from outside the function scope, either directly or through "
        "function parameters, try instantiating the object inside the decorated "
        "function instead."
    )
    with pytest.raises(RuntimeError, match=serialization_failed):
        fails_serialization()


def test_serialization_wrapping():
    def my_entry(*args, **kwargs):
        print("something with \" and ' and \n")
        return args, kwargs

    args, kwargs = (1, "two"), {"three": 3}
    template = _serialize_entry_point(my_entry, args, kwargs)
    pickled_str = re.search(r"(?s)cloudpickle.loads\((.*?)\)\ndef my_entry", template).group(1)
    byte_str = ast.literal_eval(pickled_str)

    recovered = cloudpickle.loads(byte_str)
    assert recovered() == (args, kwargs)


def test_python_validation(aws_session):
    aws_session.get_full_image_tag.return_value = "1.0-cpu-py38-ubuntu22.04"

    bad_version = (
        "Python version must match between local environment and container. "
        f"Client is running Python {sys.version_info.major}.{sys.version_info.minor} "
        "locally, but container uses Python 3.8."
    )
    with pytest.raises(RuntimeError, match=bad_version):

        @hybrid_job(device=None, aws_session=aws_session)
        def my_job():
            pass


@pytest.mark.parametrize(
    "hyperparameter, expected",
    (
        (
            "with\nnewline",
            "with newline",
        ),
        (
            "with weird chars: (&$`)",
            "with weird chars: {+?'}",
        ),
        (
            "?" * 2600,
            f"{'?' * 2477}...{'?' * 20}",
        ),
    ),
)
def test_sanitize_hyperparameters(hyperparameter, expected):
    assert _sanitize(hyperparameter) == expected



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json

import numpy as np
import pytest

import braket.ir.jaqcd as ir
from braket.circuits import Circuit, Instruction, Noise, QubitSet
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)
from braket.ir.jaqcd.shared_models import (
    DampingProbability,
    DampingSingleProbability,
    DoubleControl,
    DoubleTarget,
    MultiProbability,
    MultiTarget,
    SingleControl,
    SingleProbability,
    SingleProbability_34,
    SingleProbability_1516,
    SingleTarget,
    TripleProbability,
    TwoDimensionalMatrixList,
)

testdata = [
    (Noise.BitFlip, "bit_flip", ir.BitFlip, [SingleTarget, SingleProbability], {}),
    (Noise.PhaseFlip, "phase_flip", ir.PhaseFlip, [SingleTarget, SingleProbability], {}),
    (Noise.Depolarizing, "depolarizing", ir.Depolarizing, [SingleTarget, SingleProbability_34], {}),
    (
        Noise.AmplitudeDamping,
        "amplitude_damping",
        ir.AmplitudeDamping,
        [SingleTarget, DampingProbability],
        {},
    ),
    (
        Noise.GeneralizedAmplitudeDamping,
        "generalized_amplitude_damping",
        ir.GeneralizedAmplitudeDamping,
        [SingleTarget, DampingProbability, DampingSingleProbability],
        {},
    ),
    (
        Noise.PhaseDamping,
        "phase_damping",
        ir.PhaseDamping,
        [SingleTarget, DampingProbability],
        {},
    ),
    (
        Noise.TwoQubitDepolarizing,
        "two_qubit_depolarizing",
        ir.TwoQubitDepolarizing,
        [DoubleTarget, SingleProbability_1516],
        {},
    ),
    (
        Noise.TwoQubitDephasing,
        "two_qubit_dephasing",
        ir.TwoQubitDephasing,
        [DoubleTarget, SingleProbability_34],
        {},
    ),
    (
        Noise.TwoQubitPauliChannel,
        "two_qubit_pauli_channel",
        ir.MultiQubitPauliChannel,
        [DoubleTarget, MultiProbability],
        {},
    ),
    (
        Noise.PauliChannel,
        "pauli_channel",
        ir.PauliChannel,
        [SingleTarget, TripleProbability],
        {},
    ),
    (
        Noise.Kraus,
        "kraus",
        ir.Kraus,
        [TwoDimensionalMatrixList, MultiTarget],
        {"input_type": complex},
    ),
    (
        Noise.Kraus,
        "kraus",
        ir.Kraus,
        [TwoDimensionalMatrixList, MultiTarget],
        {"input_type": float},
    ),
    (
        Noise.Kraus,
        "kraus",
        ir.Kraus,
        [TwoDimensionalMatrixList, MultiTarget],
        {"input_type": int},
    ),
]


invalid_kraus_matrices = [
    ([np.array([[1]])]),
    ([np.array([1])]),
    ([np.array([0, 1, 2])]),
    ([np.array([[0, 1], [1, 2], [3, 4]])]),
    ([np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])]),
    ([np.array([[0, 1], [1, 1]])]),
    ([np.array([[1, 0], [0, 1]]), np.array([[0, 1], [1, 0]])]),
    ([np.array([[1, 0], [0, 1]]) * np.sqrt(0.5), np.eye(4) * np.sqrt(0.5)]),
    ([np.eye(8)]),
    ([np.eye(2), np.eye(2), np.eye(2), np.eye(2), np.eye(2)]),
]


def single_target_valid_input(**kwargs):
    return {"target": 2}


def double_target_valid_ir_input(**kwargs):
    return {"targets": [2, 3]}


def double_target_valid_input(**kwargs):
    return {"target1": 2, "target2": 3}


def single_probability_valid_input(**kwargs):
    return {"probability": 0.1234}


def single_probability_34_valid_input(**kwargs):
    return {"probability": 0.1234}


def single_probability_1516_valid_input(**kwargs):
    return {"probability": 0.1234}


def damping_single_probability_valid_input(**kwargs):
    return {"probability": 0.1234}


def damping_probability_valid_input(**kwargs):
    return {"gamma": 0.1234}


def triple_probability_valid_input(**kwargs):
    return {"probX": 0.1234, "probY": 0.1324, "probZ": 0.1423}


def single_control_valid_input(**kwargs):
    return {"control": 0}


def double_control_valid_ir_input(**kwargs):
    return {"controls": [0, 1]}


def double_control_valid_input(**kwargs):
    return {"control1": 0, "control2": 1}


def multi_target_valid_input(**kwargs):
    return {"targets": [5]}


def two_dimensional_matrix_list_valid_ir_input(**kwargs):
    return {"matrices": [[[[0, 0], [1, 0]], [[1, 0], [0, 0]]]]}


def two_dimensional_matrix_list_valid_input(**kwargs):
    input_type = kwargs.get("input_type")
    return {
        "matrices": [np.array([[input_type(0), input_type(1)], [input_type(1), input_type(0)]])]
    }


def multi_probability_valid_input(**kwargs):
    return {"probabilities": {"XX": 0.1}}


def multi_probability_invalid_input(**kwargs):
    return {"probabilities": {"XX": 1.1}}


valid_ir_switcher = {
    "SingleTarget": single_target_valid_input,
    "DoubleTarget": double_target_valid_ir_input,
    "SingleProbability": single_probability_valid_input,
    "SingleProbability_34": single_probability_34_valid_input,
    "SingleProbability_1516": single_probability_1516_valid_input,
    "DampingProbability": damping_probability_valid_input,
    "DampingSingleProbability": damping_single_probability_valid_input,
    "TripleProbability": triple_probability_valid_input,
    "MultiProbability": multi_probability_valid_input,
    "SingleControl": single_control_valid_input,
    "DoubleControl": double_control_valid_ir_input,
    "MultiTarget": multi_target_valid_input,
    "TwoDimensionalMatrixList": two_dimensional_matrix_list_valid_ir_input,
}


valid_subroutine_switcher = dict(
    valid_ir_switcher,
    **{
        "TwoDimensionalMatrixList": two_dimensional_matrix_list_valid_input,
        "DoubleTarget": double_target_valid_input,
        "DoubleControl": double_control_valid_input,
    },
)


def create_valid_ir_input(irsubclasses):
    input = {}
    for subclass in irsubclasses:
        input.update(valid_ir_switcher.get(subclass.__name__, lambda: "Invalid subclass")())
    return input


def create_valid_subroutine_input(irsubclasses, **kwargs):
    input = {}
    for subclass in irsubclasses:
        input.update(
            valid_subroutine_switcher.get(subclass.__name__, lambda: "Invalid subclass")(**kwargs)
        )
    return input


def create_valid_target_input(irsubclasses):
    input = {}
    qubit_set = []
    # based on the concept that control goes first in target input
    for subclass in irsubclasses:
        if subclass == SingleTarget:
            qubit_set.extend(list(single_target_valid_input().values()))
        elif subclass == DoubleTarget:
            qubit_set.extend(list(double_target_valid_ir_input().values()))
        elif subclass == MultiTarget:
            qubit_set.extend(list(multi_target_valid_input().values()))
        elif subclass == SingleControl:
            qubit_set = list(single_control_valid_input().values()) + qubit_set
        elif subclass == DoubleControl:
            qubit_set = list(double_control_valid_ir_input().values()) + qubit_set
        elif any(
            subclass == i
            for i in [
                SingleProbability,
                SingleProbability_34,
                SingleProbability_1516,
                DampingSingleProbability,
                DampingProbability,
                TripleProbability,
                TwoDimensionalMatrixList,
                MultiProbability,
            ]
        ):
            pass
        else:
            raise ValueError("Invalid subclass")
    input["target"] = QubitSet(qubit_set)
    return input


def create_valid_noise_class_input(irsubclasses, **kwargs):
    input = {}
    if SingleProbability in irsubclasses:
        input.update(single_probability_valid_input())
    if SingleProbability_34 in irsubclasses:
        input.update(single_probability_34_valid_input())
    if SingleProbability_1516 in irsubclasses:
        input.update(single_probability_1516_valid_input())
    if DampingSingleProbability in irsubclasses:
        input.update(damping_single_probability_valid_input())
    if DampingProbability in irsubclasses:
        input.update(damping_probability_valid_input())
    if TripleProbability in irsubclasses:
        input.update(triple_probability_valid_input())
    if MultiProbability in irsubclasses:
        input.update(multi_probability_valid_input())
    if TwoDimensionalMatrixList in irsubclasses:
        input.update(two_dimensional_matrix_list_valid_input(**kwargs))
    return input


def create_valid_instruction_input(testclass, irsubclasses, **kwargs):
    input = create_valid_target_input(irsubclasses)
    input["operator"] = testclass(**create_valid_noise_class_input(irsubclasses, **kwargs))
    return input


def calculate_qubit_count(irsubclasses):
    qubit_count = 0
    for subclass in irsubclasses:
        if subclass == SingleTarget:
            qubit_count += 1
        elif subclass == DoubleTarget:
            qubit_count += 2
        elif subclass == SingleControl:
            qubit_count += 1
        elif subclass == DoubleControl:
            qubit_count += 2
        elif subclass == MultiTarget:
            qubit_count += 3
        elif any(
            subclass == i
            for i in [
                SingleProbability,
                SingleProbability_34,
                SingleProbability_1516,
                DampingSingleProbability,
                DampingProbability,
                TripleProbability,
                MultiProbability,
                TwoDimensionalMatrixList,
            ]
        ):
            pass
        else:
            raise ValueError("Invalid subclass")
    return qubit_count


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_noise_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    expected = irclass(**create_valid_ir_input(irsubclasses))
    actual = testclass(**create_valid_noise_class_input(irsubclasses, **kwargs)).to_ir(
        **create_valid_target_input(irsubclasses)
    )
    assert actual == expected


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_instruction_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    expected = irclass(**create_valid_ir_input(irsubclasses))
    instruction = Instruction(**create_valid_instruction_input(testclass, irsubclasses, **kwargs))
    actual = instruction.to_ir()
    assert actual == expected


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_noise_subroutine(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    qubit_count = calculate_qubit_count(irsubclasses)
    subroutine = getattr(Circuit(), subroutine_name)
    assert subroutine(**create_valid_subroutine_input(irsubclasses, **kwargs)) == Circuit(
        Instruction(**create_valid_instruction_input(testclass, irsubclasses, **kwargs))
    )
    if qubit_count == 1:
        multi_targets = [0, 1, 2]
        instruction_list = []
        for target in multi_targets:
            instruction_list.append(
                Instruction(
                    operator=testclass(**create_valid_noise_class_input(irsubclasses, **kwargs)),
                    target=target,
                )
            )
        subroutine = getattr(Circuit(), subroutine_name)
        subroutine_input = {"target": multi_targets}
        if SingleProbability in irsubclasses:
            subroutine_input.update(single_probability_valid_input())
        if SingleProbability_34 in irsubclasses:
            subroutine_input.update(single_probability_34_valid_input())
        if SingleProbability_1516 in irsubclasses:
            subroutine_input.update(single_probability_1516_valid_input())
        if DampingSingleProbability in irsubclasses:
            subroutine_input.update(damping_single_probability_valid_input())
        if DampingProbability in irsubclasses:
            subroutine_input.update(damping_probability_valid_input())
        if TripleProbability in irsubclasses:
            subroutine_input.update(triple_probability_valid_input())
        if MultiProbability in irsubclasses:
            subroutine_input.update(multi_probability_valid_input())

        circuit1 = subroutine(**subroutine_input)
        circuit2 = Circuit(instruction_list)
        assert circuit1 == circuit2


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_noise_to_matrix(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    noise1 = testclass(**create_valid_noise_class_input(irsubclasses, **kwargs))
    noise2 = testclass(**create_valid_noise_class_input(irsubclasses, **kwargs))
    assert all(isinstance(matrix, np.ndarray) for matrix in noise1.to_matrix())
    assert all(np.allclose(m1, m2) for m1, m2 in zip(noise1.to_matrix(), noise2.to_matrix()))


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_fixed_qubit_count(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    fixed_qubit_count = testclass.fixed_qubit_count()
    if fixed_qubit_count is not NotImplemented:
        noise = testclass(**create_valid_noise_class_input(irsubclasses, **kwargs))
        assert noise.qubit_count == fixed_qubit_count


@pytest.mark.parametrize(
    "parameterized_noise",
    [
        (Noise.BitFlip(0.1)),
        (Noise.BitFlip(FreeParameter("alpha"))),
        (Noise.PhaseFlip(0.1)),
        (Noise.PhaseFlip(FreeParameter("alpha"))),
        (Noise.Depolarizing(0.1)),
        (Noise.Depolarizing(FreeParameter("alpha"))),
        (Noise.AmplitudeDamping(0.1)),
        (Noise.AmplitudeDamping(FreeParameter("alpha"))),
        (Noise.GeneralizedAmplitudeDamping(0.1, 0.2)),
        (Noise.GeneralizedAmplitudeDamping(FreeParameter("alpha"), FreeParameter("beta"))),
        (Noise.PhaseDamping(0.1)),
        (Noise.PhaseDamping(FreeParameter("alpha"))),
        (Noise.TwoQubitDepolarizing(0.1)),
        (Noise.TwoQubitDepolarizing(FreeParameter("alpha"))),
        (Noise.TwoQubitDephasing(0.1)),
        (Noise.TwoQubitDephasing(FreeParameter("alpha"))),
        (Noise.TwoQubitPauliChannel({"XX": 0.1, "YY": 0.2})),
        (Noise.TwoQubitPauliChannel({"XX": FreeParameter("x"), "YY": FreeParameter("y")})),
        (Noise.PauliChannel(0.1, 0.2, 0.3)),
        (Noise.PauliChannel(FreeParameter("x"), FreeParameter("y"), FreeParameter("z"))),
    ],
)
def test_serialization(parameterized_noise):
    serialized = parameterized_noise.to_dict()
    serialized_str = json.dumps(serialized)
    deserialized_dict = json.loads(serialized_str)
    deserialized = Noise.from_dict(deserialized_dict)
    assert deserialized == parameterized_noise


@pytest.mark.parametrize(
    "parameterized_noise, params, expected_noise",
    [
        (Noise.BitFlip(FreeParameter("alpha")), {"alpha": 0.1}, Noise.BitFlip(0.1)),
        (Noise.PhaseFlip(FreeParameter("alpha")), {"alpha": 0.1}, Noise.PhaseFlip(0.1)),
        (Noise.Depolarizing(FreeParameter("alpha")), {"alpha": 0.1}, Noise.Depolarizing(0.1)),
        (
            Noise.AmplitudeDamping(FreeParameter("alpha")),
            {"alpha": 0.1},
            Noise.AmplitudeDamping(0.1),
        ),
        (
            Noise.GeneralizedAmplitudeDamping(FreeParameter("alpha"), FreeParameter("beta")),
            {"alpha": 0.1},
            Noise.GeneralizedAmplitudeDamping(0.1, FreeParameter("beta")),
        ),
        (Noise.PhaseDamping(FreeParameter("alpha")), {"alpha": 0.1}, Noise.PhaseDamping(0.1)),
        (
            Noise.TwoQubitDepolarizing(FreeParameter("alpha")),
            {"alpha": 0.1},
            Noise.TwoQubitDepolarizing(0.1),
        ),
        (
            Noise.TwoQubitDephasing(FreeParameter("alpha")),
            {"alpha": 0.1},
            Noise.TwoQubitDephasing(0.1),
        ),
        (
            Noise.TwoQubitPauliChannel({"XX": FreeParameter("x"), "YY": FreeParameter("y")}),
            {"x": 0.1},
            Noise.TwoQubitPauliChannel({"XX": 0.1, "YY": FreeParameter("y")}),
        ),
        (
            Noise.PauliChannel(FreeParameter("x"), FreeParameter("y"), FreeParameter("z")),
            {"x": 0.1, "z": 0.2},
            Noise.PauliChannel(0.1, FreeParameter("y"), 0.2),
        ),
    ],
)
def test_parameter_binding(parameterized_noise, params, expected_noise):
    result_noise = parameterized_noise.bind_values(**params)
    assert result_noise == expected_noise


def test_parameterized_noise():
    noise = Noise.PauliChannel(FreeParameter("a"), 0.2, FreeParameter("b"))
    assert noise.probX == FreeParameter("a")
    assert noise.probY == 0.2
    assert noise.probZ == FreeParameter("b")


# Additional Unitary noise tests


@pytest.mark.parametrize("matrices", invalid_kraus_matrices)
def test_kraus_invalid_matrix(matrices):
    with pytest.raises(ValueError):
        Noise.Kraus(matrices=matrices)


def test_kraus_matrix_target_size_mismatch():
    with pytest.raises(ValueError):
        Circuit().kraus(
            matrices=[np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])],
            targets=[0],
        )


@pytest.mark.parametrize(
    "probs",
    [
        {"X": -0.1},
        {"XY": 1.1},
        {"TX": 0.1},
        {"X": 0.5, "Y": 0.6},
        {"X": 0.1, "YY": 0.2},
        {"II": 0.9, "XX": 0.1},
    ],
)
def test_invalid_values_pauli_channel_two_qubit(probs):
    with pytest.raises(ValueError):
        Noise.TwoQubitPauliChannel(probs)


@pytest.mark.parametrize(
    "probs",
    [
        {"XY": 0.1},
        {"XX": 0.1, "ZZ": 0.2},
    ],
)
def test_valid_values_pauli_channel_two_qubit(probs):
    noise = Noise.TwoQubitPauliChannel(probs)
    assert len(noise.to_matrix()) == 16


@pytest.mark.parametrize(
    "noise, serialization_properties, target, expected_ir",
    [
        (
            Noise.BitFlip(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise bit_flip(0.5) q[3]",
        ),
        (
            Noise.BitFlip(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise bit_flip(0.5) $3",
        ),
        (
            Noise.PhaseFlip(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise phase_flip(0.5) q[3]",
        ),
        (
            Noise.PhaseFlip(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise phase_flip(0.5) $3",
        ),
        (
            Noise.PauliChannel(0.1, 0.2, 0.3),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise pauli_channel(0.1, 0.2, 0.3) q[3]",
        ),
        (
            Noise.PauliChannel(0.1, 0.2, 0.3),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise pauli_channel(0.1, 0.2, 0.3) $3",
        ),
        (
            Noise.Depolarizing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise depolarizing(0.5) q[3]",
        ),
        (
            Noise.Depolarizing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise depolarizing(0.5) $3",
        ),
        (
            Noise.TwoQubitDepolarizing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 5],
            "#pragma braket noise two_qubit_depolarizing(0.5) q[3], q[5]",
        ),
        (
            Noise.TwoQubitDepolarizing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 5],
            "#pragma braket noise two_qubit_depolarizing(0.5) $3, $5",
        ),
        (
            Noise.TwoQubitDephasing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 5],
            "#pragma braket noise two_qubit_dephasing(0.5) q[3], q[5]",
        ),
        (
            Noise.TwoQubitDephasing(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 5],
            "#pragma braket noise two_qubit_dephasing(0.5) $3, $5",
        ),
        (
            Noise.AmplitudeDamping(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise amplitude_damping(0.5) q[3]",
        ),
        (
            Noise.AmplitudeDamping(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise amplitude_damping(0.5) $3",
        ),
        (
            Noise.GeneralizedAmplitudeDamping(0.5, 0.1),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise generalized_amplitude_damping(0.5, 0.1) q[3]",
        ),
        (
            Noise.GeneralizedAmplitudeDamping(0.5, 0.1),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise generalized_amplitude_damping(0.5, 0.1) $3",
        ),
        (
            Noise.PhaseDamping(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise phase_damping(0.5) q[3]",
        ),
        (
            Noise.PhaseDamping(0.5),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise phase_damping(0.5) $3",
        ),
        (
            Noise.Kraus(
                [
                    np.eye(4) * np.sqrt(0.9),
                    np.kron([[1.0, 0.0], [0.0, 1.0]], [[0.0, 1.0], [1.0, 0.0]]) * np.sqrt(0.1),
                ]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3, 5],
            "#pragma braket noise kraus(["
            "[0.9486832980505138, 0, 0, 0], "
            "[0, 0.9486832980505138, 0, 0], "
            "[0, 0, 0.9486832980505138, 0], "
            "[0, 0, 0, 0.9486832980505138]], ["
            "[0, 0.31622776601683794, 0, 0], "
            "[0.31622776601683794, 0, 0, 0], "
            "[0, 0, 0, 0.31622776601683794], "
            "[0, 0, 0.31622776601683794, 0]]) q[3], q[5]",
        ),
        (
            Noise.Kraus(
                [
                    np.eye(4) * np.sqrt(0.9),
                    np.kron([[1.0, 0.0], [0.0, 1.0]], [[0.0, 1.0], [1.0, 0.0]]) * np.sqrt(0.1),
                ]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3, 5],
            "#pragma braket noise kraus(["
            "[0.9486832980505138, 0, 0, 0], "
            "[0, 0.9486832980505138, 0, 0], "
            "[0, 0, 0.9486832980505138, 0], "
            "[0, 0, 0, 0.9486832980505138]], ["
            "[0, 0.31622776601683794, 0, 0], "
            "[0.31622776601683794, 0, 0, 0], "
            "[0, 0, 0, 0.31622776601683794], "
            "[0, 0, 0.31622776601683794, 0]]) $3, $5",
        ),
        (
            Noise.Kraus(
                [
                    np.array([[0.9486833j, 0], [0, 0.9486833j]]),
                    np.array([[0, 0.31622777], [0.31622777, 0]]),
                ]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            [3],
            "#pragma braket noise kraus(["
            "[0.9486833im, 0], [0, 0.9486833im]], ["
            "[0, 0.31622777], [0.31622777, 0]]) q[3]",
        ),
        (
            Noise.Kraus(
                [
                    np.array([[0.9486833j, 0], [0, 0.9486833j]]),
                    np.array([[0, 0.31622777], [0.31622777, 0]]),
                ]
            ),
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            [3],
            "#pragma braket noise kraus(["
            "[0.9486833im, 0], [0, 0.9486833im]], ["
            "[0, 0.31622777], [0.31622777, 0]]) $3",
        ),
    ],
)
def test_noise_to_ir_openqasm(noise, serialization_properties, target, expected_ir):
    assert (
        noise.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )
        == expected_ir
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import os
import re
import tempfile
from pathlib import Path

import pytest

from braket.jobs.local import LocalQuantumJob


def test_completed_local_job(aws_session, capsys):
    """Asserts the hybrid job is completed with the respective files and folders for logs,
    results and checkpoints. Validate the results are what we expect. Also,
    assert that logs contains all the necessary steps for setup and running
    the hybrid job is displayed to the user.
    """
    absolute_source_module = str(Path("test/integ_tests/job_test_script.py").resolve())
    current_dir = Path.cwd()

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            os.chdir(temp_dir)
            job = LocalQuantumJob.create(
                "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
                source_module=absolute_source_module,
                entry_point="job_test_script:start_here",
                hyperparameters={"test_case": "completed"},
                aws_session=aws_session,
            )

            job_name = job.name
            pattern = f"^local:job/{job_name}$"
            re.match(pattern=pattern, string=job.arn)

            assert job.state() == "COMPLETED"
            assert Path(job_name).is_dir()

            # Check results match the expectations.
            assert Path(f"{job_name}/results.json").exists()
            assert job.result() == {"converged": True, "energy": -0.2}

            # Validate checkpoint files and data
            assert Path(f"{job_name}/checkpoints/{job_name}.json").exists()
            assert Path(f"{job_name}/checkpoints/{job_name}_plain_data.json").exists()

            for file_name, expected_data in [
                (
                    f"{job_name}/checkpoints/{job_name}_plain_data.json",
                    {
                        "braketSchemaHeader": {
                            "name": "braket.jobs_data.persisted_job_data",
                            "version": "1",
                        },
                        "dataDictionary": {"some_data": "abc"},
                        "dataFormat": "plaintext",
                    },
                ),
                (
                    f"{job_name}/checkpoints/{job_name}.json",
                    {
                        "braketSchemaHeader": {
                            "name": "braket.jobs_data.persisted_job_data",
                            "version": "1",
                        },
                        "dataDictionary": {"some_data": "gASVBwAAAAAAAACMA2FiY5Qu\n"},
                        "dataFormat": "pickled_v4",
                    },
                ),
            ]:
                with open(file_name, "r") as f:
                    assert json.loads(f.read()) == expected_data

            # Capture logs
            assert Path(f"{job_name}/log.txt").exists()
            job.logs()
            log_data, errors = capsys.readouterr()

            logs_to_validate = [
                "Beginning Setup",
                "Running Code As Process",
                "Test job started!!!!!",
                "Test job completed!!!!!",
                "Code Run Finished",
            ]

            for data in logs_to_validate:
                assert data in log_data

        finally:
            os.chdir(current_dir)


def test_failed_local_job(aws_session, capsys):
    """Asserts the hybrid job is failed with the output, checkpoints not created in bucket
    and only logs are populated. Validate the calling result function raises
    the ValueError. Also, check if the logs displays the required error message.
    """
    absolute_source_module = str(Path("test/integ_tests/job_test_script.py").resolve())
    current_dir = Path.cwd()

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            os.chdir(temp_dir)
            job = LocalQuantumJob.create(
                "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
                source_module=absolute_source_module,
                entry_point="job_test_script:start_here",
                hyperparameters={"test_case": "failed"},
                aws_session=aws_session,
            )

            job_name = job.name
            pattern = f"^local:job/{job_name}$"
            re.match(pattern=pattern, string=job.arn)

            assert Path(job_name).is_dir()

            # Check no files are populated in checkpoints folder.
            assert not any(Path(f"{job_name}/checkpoints").iterdir())

            # Check results match the expectations.
            error_message = f"Unable to find results in the local job directory {job_name}."
            with pytest.raises(ValueError, match=error_message):
                job.result()

            assert Path(f"{job_name}/log.txt").exists()
            job.logs()
            log_data, errors = capsys.readouterr()

            logs_to_validate = [
                "Beginning Setup",
                "Running Code As Process",
                "Test job started!!!!!",
                "Code Run Finished",
            ]

            for data in logs_to_validate:
                assert data in log_data
        finally:
            os.chdir(current_dir)
